"""Module providing various functions for advanced handling of pandas dataframes."""

import re
from collections.abc import Sequence
from typing import Callable, Optional, Union

import numpy as np
import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_has_columns, _assert_has_index_levels, _assert_is_dtype
from biopsykit.utils.dtypes import CodebookDataFrame, is_codebook_dataframe

__all__ = [
    "add_space_to_camel",
    "apply_codebook",
    "camel_to_snake",
    "convert_nan",
    "int_from_str_col",
    "int_from_str_idx",
    "multi_xs",
    "replace_missing_data",
    "snake_to_camel",
    "stack_groups_percent",
    "wide_to_long",
]


def int_from_str_idx(
    data: pd.DataFrame,
    idx_levels: Union[str, Sequence[str]],
    regex: Union[str, Sequence[str]],
    func: Optional[Callable] = None,
) -> pd.DataFrame:
    """Extract integers from strings in index levels and set them as new index values.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data with index to extract information from
    idx_levels : str or list of str
        name of index level or list of index level names
    regex : str or list of str
        regex string or list of regex strings to extract integers from strings
    func : function, optional
        function to apply to the extracted integer values. This can, for example, be a lambda function which
        increments all integers by 1. Default: ``None``


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with new index values

    """
    if isinstance(idx_levels, str):
        idx_levels = [idx_levels]
    if isinstance(regex, str):
        regex = [regex]

    if len(idx_levels) != len(regex):
        raise ValueError(
            "Number of values in 'regex' must match number of index levels in 'idx_levels'! "
            f"Got idx_levels: {idx_levels}, regex: {regex}."
        )

    _assert_is_dtype(data, pd.DataFrame)
    _assert_has_index_levels(data, idx_levels, match_atleast=True, match_order=False)

    idx_names = data.index.names
    data = data.reset_index()
    for idx, reg in zip(idx_levels, regex):
        idx_col = data[idx].str.extract(reg).astype(int)[0]
        if func is not None:
            idx_col = func(idx_col)
        data[idx] = idx_col

    data = data.set_index(idx_names)
    return data


def int_from_str_col(
    data: pd.DataFrame, col_name: str, regex: str, func: Optional[Callable] = None
) -> Union[pd.Series]:
    """Extract integers from strings in the column of a dataframe and return it.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data with column names to extract information from
    col_name : str
        name of column with string values to extract
    regex : str
        regex string used to extract integers from string values
    func : function, optional
        function to apply to the extracted integer values. This can, for example, be a lambda function which
        increments all integers by 1. Default: ``None``


    Returns
    -------
    :class:`~pandas.Series`
        series object with extracted integer values

    """
    _assert_is_dtype(data, pd.DataFrame)
    _assert_has_columns(data, [[col_name]])

    column = data[col_name].str.extract(regex).astype(int)[0]
    if func is not None:
        column = func(column)
    return column


def camel_to_snake(name: str, lower: Optional[bool] = True):
    """Convert string in "camelCase" to "snake_case".

    .. note::
        If all letters in ``name`` are capital letters the string will not be computed into snake_case because
        it is assumed to be an abbreviation.

    Parameters
    ----------
    name : str
        string to convert from camelCase to snake_case
    lower : bool, optional
        ``True`` to convert all capital letters in to lower case ("actual" snake_case), ``False`` to keep
        capital letters, if present


    Returns
    -------
    str
        string converted into snake_case

    Examples
    --------
    >>> from biopsykit.utils.dataframe_handling import camel_to_snake
    >>> camel_to_snake("HelloWorld")
    hello_world
    >>> camel_to_snake("HelloWorld", lower=False)
    Hello_World
    >>> camel_to_snake("ABC")
    ABC

    """
    if not name.isupper():
        name = re.sub(r"(?<!^)(?=[A-Z])", "_", name)
        if lower:
            name = name.lower()
    return name


def snake_to_camel(name: str):
    """Convert string in "snake_case" to "camelCase".

    Parameters
    ----------
    name : str
        string to convert from snake_case to camelCase

    """
    return "".join(x.capitalize() or "_" for x in name.split("_"))


def add_space_to_camel(name: str) -> str:
    """Add space to string in "camelCase".

    Parameters
    ----------
    name : str
        string to transform

    Returns
    -------
    str
        string with space added

    Examples
    --------
    >>> from biopsykit.utils.dataframe_handling import add_space_to_camel
    >>> add_space_to_camel("HelloWorld")
    Hello World
    >>> add_space_to_camel("ABC")
    ABC

    """
    return re.sub(r"((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))", r" \1", name)


def replace_missing_data(
    data: pd.DataFrame,
    target_col: str,
    source_col: str,
    dropna: Optional[bool] = False,
    inplace: Optional[bool] = False,
) -> Optional[pd.DataFrame]:
    """Replace missing data in one column by data from another column.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data with values to replace
    target_col : str
        target column, i.e., column in which missing values should be replaced
    source_col : str
        source column, i.e., column values used to replace missing values in ``target_col``
    dropna : bool, optional
        whether to drop rows with missing values in ``target_col`` or not. Default: ``False``
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame` or ``None``
        dataframe with replaced missing values or ``None`` if ``inplace`` is ``True``

    """
    _assert_is_dtype(data, pd.DataFrame)
    if not inplace:
        data = data.copy()

    data[target_col].fillna(data[source_col], inplace=True)  # noqa: PD002
    if dropna:
        data = data.dropna(subset=[target_col])

    if inplace:
        return None
    return data


def convert_nan(
    data: Union[pd.DataFrame, pd.Series], inplace: Optional[bool] = False
) -> Union[pd.DataFrame, pd.Series, None]:
    """Convert missing values to NaN.

    Data exported from programs like SPSS often uses negative integers to encode missing values because these negative
    numbers are "unrealistic" values. Use this function to convert these negative numbers to
    "actual" missing values: not-a-number (``NaN``).

    Values that will be replaced with ``NaN`` are -66, -77, -99 (integer and string representations).

    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        input data
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~pandas.DataFrame` or ``None``
        dataframe with converted missing values or ``None`` if ``inplace`` is ``True``

    """
    _assert_is_dtype(data, (pd.DataFrame, pd.Series))

    if not inplace:
        data = data.copy()
    data = data.replace([-99.0, -77.0, -66.0, "-99", "-77", "-66"], np.nan)
    if inplace:
        return None
    return data


def multi_xs(
    data: Union[pd.DataFrame, pd.Series],
    keys: Union[str, Sequence[str]],
    axis: int = 0,
    level: Optional[Union[str, int, Sequence[str], Sequence[int]]] = None,
    drop_level: Optional[bool] = True,
) -> Union[pd.DataFrame, pd.Series]:
    """Return cross-section of multiple keys from the dataframe.

    This function internally calls the :meth:`pandas.DataFrame.xs` method, but it can take a list of key arguments
    to return multiple keys at once, in comparison to the original :meth:`~pandas.DataFrame.xs` method which
    only takes one possible key.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        input data to get cross-section from
    keys : str or list of str
        label(s) contained in the index, or partially in a :class:`~pandas.MultiIndex`
    level : str, int, or list of such
        in case of keys partially contained in a :class:`~pandas.MultiIndex`, indicate which index levels are used.
        Levels can be referred by label or position.
    drop_level : bool, optional
        if ``False``, returns object with same levels as self. Default: ``True``


    Returns
    -------
    :class:`~pandas.DataFrame` or :class:`~pandas.Series`
        cross-section from the original dataframe or series

    """
    _assert_is_dtype(data, (pd.DataFrame, pd.Series))
    if isinstance(keys, str):
        keys = [keys]

    level_concat = level
    if isinstance(level, (str, int)):
        level_concat = [level_concat]
    levels = data.columns.names if axis == 1 else data.index.names

    data_xs = pd.concat(
        {key: data.xs(key, level=level, drop_level=drop_level, axis=axis) for key in keys},
        names=level_concat,
        axis=axis,
    )

    return data_xs.reorder_levels(levels, axis=axis).sort_index(axis=axis)


def stack_groups_percent(
    data: pd.DataFrame, hue: str, stacked: str, order: Optional[Sequence[str]] = None
) -> pd.DataFrame:
    """Create dataframe with stacked groups.

    To create a stacked bar chart, i.e. a plot with different bar charts along a categorical axis,
    where the variables of each bar chart are stacked along the value axis, the data needs to be rearranged and
    normalized in percent.

    The columns of the resulting dataframe be the categorical values specified by ``hue``,
    the index items will be the variables specified by ``stacked``.


    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to compute stacked group in percent
    hue : str
        column name of grouping categorical variable. This typically corresponds to the ``x`` axis
        in a stacked bar chart.
    stacked : str
        column name of variable that is stacked along the ``y`` axis
    order : str
        order of categorical variable specified by ``hue``


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe in a format that can be used to create a stacked bar chart

    See Also
    --------
    :func:`~biopsykit.plotting.stacked_barchart`
        function to create a stacked bar chart

    """
    data_grouped = pd.DataFrame(data.groupby([hue, stacked]).size(), columns=["data"])
    data_grouped = data_grouped.groupby(hue).apply(lambda x: 100 * (x / x.sum())).T.stack().T
    if order is not None:
        data_grouped = data_grouped.reindex(order)
    return data_grouped["data"]


def apply_codebook(data: pd.DataFrame, codebook: CodebookDataFrame) -> pd.DataFrame:
    """Apply codebook to convert numerical to categorical values.

    The codebook is expected to be a dataframe in a standardized format
    (see :obj:`~biopsykit.utils.dtypes.CodebookDataFrame` for further information).



    Parameters
    ----------
    codebook : :obj:`~biopsykit.utils.dtypes.CodebookDataFrame`
        path to codebook or dataframe to be used as codebook
    data : :class:`~pandas.DataFrame`
        data to apply codebook on

    Returns
    -------
    :class:`~pandas.DataFrame`
        data with numerical values converted to categorical values

    See Also
    --------
    :func:`~biopsykit.io.load_codebook`
        load Codebook

    Examples
    --------
    >>> codebook = pd.DataFrame(
    >>>     {
    >>>         0: [None, None, "Morning"],
    >>>         1: ["Male", "No", "Intermediate"],
    >>>         2: ["Female", "Not very often", "Evening"],
    >>>         3: [None, "Often", None],
    >>>         4: [None, "Very often", None]
    >>>     },
    >>>     index=pd.Index(["gender", "smoking", "chronotype"], name="variable")
    >>> )
    >>> apply_codebook(codebook, data)

    """
    is_codebook_dataframe(codebook)

    for col in data.index.names:
        if col in codebook.index:
            data = data.rename(index=codebook.loc[col], level=col)

    for col in data.columns:
        if col in codebook.index:
            data.loc[:, col].replace(codebook.loc[col], inplace=True)  # noqa: PD002

    return data


def wide_to_long(
    data: pd.DataFrame,
    stubname: str,
    levels: Union[str, Sequence[str]],
    sep: Optional[str] = "_",
) -> pd.DataFrame:
    """Convert a dataframe wide-format into long-format.

    In the wide-format dataframe, the index levels to be converted into long-format are expected to be encoded in the
    column names and separated by ``sep``. If multiple levels should be converted into long-format, e.g., for a
    questionnaire with subscales (level `subscale`) that was assessed pre and post (level `time`), then the different
    levels are all encoded into the string. The level order is specified by ``levels``.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        pandas DataFrame containing saliva data in wide-format, i.e. one column per saliva sample, one row per subject
    stubname : str
        common name for each column to be converted into long-format. Usually, this is either the name of the
        questionnaire (e.g., "PSS") or the saliva type (e.g., "cortisol").
    levels : str or list of str
        index levels of the resulting long-format dataframe.
    sep : str, optional
        character separating index levels in the column names of the wide-format dataframe. Default: ``_``


    Returns
    -------
    :class:`~pandas.DataFrame`
        pandas DataFrame in long-format


    Examples
    --------
    >>> data = pd.DataFrame(
    >>>     columns=[
    >>>         "MDBF_GoodBad_pre", "MDBF_AwakeTired_pre", "MDBF_CalmNervous_pre",
    >>>         "MDBF_GoodBad_post", "MDBF_AwakeTired_post",  "MDBF_CalmNervous_post"
    >>>     ],
    >>>     index=pd.Index(range(0, 5), name="subject")
    >>> )
    >>> data_long = wide_to_long(data, stubname="MDBF", levels=["subscale", "time"], sep="_")
    >>> print(data_long.index.names)
    ['subject', 'subscale', 'time']
    >>> print(data_long.index)
    MultiIndex([(0,  'AwakeTired', 'post'),
            (0,  'AwakeTired',  'pre'),
            (0, 'CalmNervous', 'post'),
            (0, 'CalmNervous',  'pre'),
            (0,     'GoodBad', 'post'),
            (0,     'GoodBad',  'pre'),
            (1,  'AwakeTired', 'post'),
            ...



    """
    if isinstance(levels, str):
        levels = [levels]

    data = data.filter(like=stubname)
    index_cols = list(data.index.names)

    if any(col is None for col in index_cols):
        raise ValueError(
            "All index levels of the dataframe need to have names! Please assign names using "
            "'pandas.Index.set_names()' before using this function!"
        )

    # reverse level order because nested multi-level index will be constructed from back to front
    levels = levels[::-1]
    # iteratively build up long-format dataframe
    for i, level in enumerate(levels):
        stubnames = list(data.columns)
        # stubnames are everything except the last part separated by underscore
        stubnames = sorted({"_".join(s.split("_")[:-1]) for s in stubnames})
        data = pd.wide_to_long(
            data.reset_index(),
            stubnames=stubnames,
            i=index_cols + levels[0:i],
            j=level,
            sep=sep,
            suffix=r"\w+",
        )

    # reorder levels and sort
    return data.reorder_levels(index_cols + levels[::-1]).sort_index()



"""Module for outlier correction of ICG signals."""

from biopsykit.signals.icg.outlier_correction._base_outlier_correction import BaseOutlierCorrection
from biopsykit.signals.icg.outlier_correction._outlier_correction_dummy import OutlierCorrectionDummy
from biopsykit.signals.icg.outlier_correction._outlier_correction_forouzanfar2018 import (
    OutlierCorrectionForouzanfar2018,
)
from biopsykit.signals.icg.outlier_correction._outlier_correction_linear_interpolation import (
    OutlierCorrectionLinearInterpolation,
)

__all__ = [
    "BaseOutlierCorrection",
    "OutlierCorrectionDummy",
    "OutlierCorrectionForouzanfar2018",
    "OutlierCorrectionLinearInterpolation",
]



"""Functions for computing sleep endpoints, i.e., parameters that characterize a recording during a sleep study."""

from collections.abc import Sequence
from numbers import Number
from typing import Optional, Union

import numpy as np
import pandas as pd

from biopsykit.utils.dtypes import SleepEndpointDataFrame, SleepEndpointDict, _SleepEndpointDataFrame


def compute_sleep_endpoints(
    sleep_wake: pd.DataFrame, bed_interval: Sequence[Union[str, int, np.datetime64]]
) -> SleepEndpointDict:
    """Compute a set of sleep endpoints based on sleep/wake information and time spent in bed.

    This functions computes the following sleep endpoints:

    * ``date``: date of recording if input data is time-aware, ``0`` otherwise. **NOTE**: If the participant went
      to bed between 12 am and 6 am (i.e, the beginning of ``bed_interval`` between 12 am and 6 am) ``date`` will
      be set to the day before (because this night is assumed to "belong" to the day before).
    * ``sleep_onset``: Sleep Onset, i.e., time of falling asleep, in absolute time
    * ``wake_onset``: Wake Onset, i.e., time of awakening, in absolute time
    * ``total_sleep_duration``: Total duration spent sleeping, i.e., the duration between the beginning of the first
      sleep interval and the end of the last sleep interval, in minutes
    * ``net_sleep_duration``: Net duration spent sleeping, in minutes
    * ``bed_interval_start``: Bed Interval Start, i.e., time when participant went to bed, in absolute time
    * ``bed_interval_end``: Bed Interval End, i.e., time when participant left bed, in absolute time
    * ``sleep_efficiency``: Sleep Efficiency, defined as the ratio between net sleep duration and total sleep duration
      in percent
    * ``sleep_onset_latency``: Sleep Onset Latency, i.e., time in bed needed to fall asleep
      (difference between *Sleep Onset* and *Bed Interval Start*), in minutes
    * ``getup_latency``: Get Up Latency, i.e., time in bed after awakening until getting up (difference between
      *Bed Interval End* and *Wake Onset*), in minutes
    * ``wake_after_sleep_onset``: Wake After Sleep Onset (WASO), i.e., total time awake after falling asleep
      (after *Sleep Onset* and before *Wake Onset*), in minutes
    * ``sleep_bouts``: List with start and end times of sleep bouts
    * ``wake_bouts``: List with start and end times of wake bouts
    * ``number_wake_bouts``: Total number of wake bouts


    Parameters
    ----------
    sleep_wake : :class:`~pandas.DataFrame`
        dataframe with sleep/wake scoring of night. 1 is expected to indicate *sleep*, 0 to indicate *wake*
    bed_interval : array_like
        beginning and end of bed interval, i.e., the time spent in bed


    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SleepEndpointDict`
        dictionary with computed sleep endpoints

    """
    # slice sleep data = data between sleep onset and wake onset during bed interval
    sleep_wake = sleep_wake.loc[bed_interval[0] : bed_interval[1]]
    # total sleep duration = length of sleep data
    total_sleep_duration = len(sleep_wake)

    # net sleep duration in minutes = length of 'sleep' predictions (value 1) in sleep data
    net_sleep_time = sleep_wake[sleep_wake["sleep_wake"].eq(1)]
    net_sleep_duration = len(net_sleep_time)
    if net_sleep_time.empty:
        return {}
    df_sw_sleep = sleep_wake[net_sleep_time.index[0] : net_sleep_time.index[-1]].copy()

    # get percent of total time asleep
    sleep_efficiency = 100.0 * (len(net_sleep_time) / len(sleep_wake))
    # wake after sleep onset = duration of wake during first and last 'sleep' sample
    wake_after_sleep_onset = len(df_sw_sleep) - int(df_sw_sleep.sum().iloc[0])

    df_sw_sleep["block"] = df_sw_sleep["sleep_wake"].diff().ne(0).cumsum()
    df_sw_sleep = df_sw_sleep.reset_index()
    df_sw_sleep = df_sw_sleep.rename(columns={"index": "time"})
    bouts = df_sw_sleep.groupby(by="block")
    df_start_stop = bouts.first()
    df_start_stop = df_start_stop.rename(columns={"time": "start"})
    df_start_stop["end"] = bouts.last()["time"]

    # add 1 min to end for continuous time coverage
    if df_start_stop["end"].dtype == np.int64:
        df_start_stop["end"] = df_start_stop["end"] + 1
    else:
        df_start_stop["end"] = df_start_stop["end"] + pd.Timedelta("1m")

    sleep_bouts = df_start_stop[df_start_stop["sleep_wake"].eq(1)].drop(columns=["sleep_wake"]).reset_index(drop=True)
    wake_bouts = df_start_stop[df_start_stop["sleep_wake"].ne(1)].drop(columns=["sleep_wake"]).reset_index(drop=True)
    num_wake_bouts = len(wake_bouts)
    sleep_onset = net_sleep_time.index[0]
    wake_onset = net_sleep_time.index[-1]

    # start and end of bed interval
    bed_start = bed_interval[0]
    bed_end = bed_interval[1]

    # sleep onset latency = duration between bed interval start and sleep onset
    sleep_onset_latency = len(sleep_wake[sleep_wake.index[0] : sleep_onset])
    # getup latency = duration between wake onset (last 'sleep' sample) and bed interval end
    getup_latency = len(sleep_wake[wake_onset : sleep_wake.index[-1]])

    if isinstance(bed_start, Number):
        date = 0
    else:
        bed_start = str(bed_start)
        bed_end = str(bed_end)
        sleep_onset = str(sleep_onset)
        wake_onset = str(wake_onset)
        date = pd.to_datetime(bed_start)
        if date.hour < 6:
            date = date - pd.Timedelta("1d")
            date = str(date.normalize())

    dict_result = {
        "date": date,
        "sleep_onset": sleep_onset,
        "wake_onset": wake_onset,
        "total_sleep_duration": total_sleep_duration,
        "net_sleep_duration": net_sleep_duration,
        "bed_interval_start": bed_start,
        "bed_interval_end": bed_end,
        "sleep_efficiency": sleep_efficiency,
        "sleep_onset_latency": sleep_onset_latency,
        "getup_latency": getup_latency,
        "wake_after_sleep_onset": wake_after_sleep_onset,
        "sleep_bouts": sleep_bouts,
        "wake_bouts": wake_bouts,
        "number_wake_bouts": num_wake_bouts,
    }
    return dict_result


def endpoints_as_df(sleep_endpoints: SleepEndpointDict) -> Optional[SleepEndpointDataFrame]:
    """Convert ``SleepEndpointDict`` into ``SleepEndpointDataFrame``.

    Parameters
    ----------
    sleep_endpoints : :obj:`~biopsykit.utils.dtypes.SleepEndpointDict`
        dictionary with computed Sleep Endpoints

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.SleepEndpointDataFrame`
        dataframe with computed Sleep Endpoints or ``None`` if ``sleep_endpoints`` is ``None``

    """
    if sleep_endpoints is None:
        return None

    sleep_endpoints = sleep_endpoints.copy()
    sleep_bouts = sleep_endpoints.pop("sleep_bouts", None).to_numpy().tolist()
    wake_bouts = sleep_endpoints.pop("wake_bouts", None).to_numpy().tolist()

    sleep_bouts = [tuple(v) for v in sleep_bouts]
    wake_bouts = [tuple(v) for v in wake_bouts]

    index = pd.to_datetime(pd.Index([sleep_endpoints["date"]], name="date"))
    sleep_endpoints.pop("date")

    df = pd.DataFrame(sleep_endpoints, index=index)
    df = df.fillna(value=np.nan)
    df["sleep_bouts"] = None
    df["wake_bouts"] = None
    df.at[df.index[0], "sleep_bouts"] = sleep_bouts  # noqa: PD008
    df.at[df.index[0], "wake_bouts"] = wake_bouts  # noqa: PD008
    return _SleepEndpointDataFrame(df)



import numpy as np
import pandas as pd

from biopsykit.signals.icg.outlier_correction._base_outlier_correction import BaseOutlierCorrection

__all__ = ["OutlierCorrectionLinearInterpolation"]


from biopsykit.utils.dtypes import BPointDataFrame, CPointDataFrame, is_b_point_dataframe, is_c_point_dataframe

# TODO add verbosity option


class OutlierCorrectionLinearInterpolation(BaseOutlierCorrection):
    """B-point outlier correction algorithm based on linear interpolation.

    This algorithm corrects outliers in B-Point data using linear interpolation.

    """

    def correct_outlier(
        self,
        *,
        b_points: BPointDataFrame,
        c_points: CPointDataFrame,
        sampling_rate_hz: float,
        **kwargs,
    ):
        """Perform outlier correction on B-Point data.

        The outliers are corrected using linear interpolation.

        The results of the outlier correction are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        b_points : :class:`~pandas.DataFrame`
            Extracted B-points. Each row contains the B-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. B-point locations can be NaN if no B-points were detected
            for certain heartbeats.
        c_points : :class:`~pandas.DataFrame`
            Extracted C-points. Each row contains the C-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. C-point locations can be NaN if no C-points were detected
            for certain heartbeats.
        sampling_rate_hz : int
            sampling rate of ICG derivative signal in hz
        kwargs : dict
            Additional keyword arguments:
                * verbose: bool, optional
                    Whether to print additional information. Default: False

        Returns
        -------
        self

        """
        verbose = kwargs.get("verbose", False)
        is_b_point_dataframe(b_points)
        is_c_point_dataframe(c_points)

        corrected_b_points = pd.DataFrame(index=b_points.index, columns=["b_point_sample"])
        # stationarize the B-Point time data
        stationary_data = self.stationarize_b_points(b_points, c_points, sampling_rate_hz)
        b_points_nan = b_points.loc[b_points["b_point_sample"].isna()]
        stationary_data.loc[b_points_nan.index, "statio_data"] = np.NaN

        # detect outliers
        outliers = self.detect_b_point_outlier(stationary_data)

        counter = 1
        if verbose:
            print(f"Detected {len(outliers)} outliers in correction cycle {counter}!")

        if len(outliers) == 0:
            is_b_point_dataframe(b_points)
            self.points_ = b_points
            return self

        # Perform the outlier correction until no more outliers are detected
        while len(outliers) > 0:
            if counter >= 200:
                break
            corrected_b_points = self._correct_outlier_linear_interpolation(
                b_points, c_points, stationary_data, outliers, stationary_data["baseline"], sampling_rate_hz
            )

            stationary_data = self.stationarize_b_points(corrected_b_points, c_points, sampling_rate_hz)
            outliers = self.detect_b_point_outlier(stationary_data)
            if verbose:
                print(f"Detected {len(outliers)} outliers in correction cycle {counter}!")
            counter += 1

        if verbose:
            print("No more outliers got detected!")

        corrected_b_points = corrected_b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        is_b_point_dataframe(corrected_b_points)

        self.points_ = corrected_b_points
        return self

    @staticmethod
    def _correct_outlier_linear_interpolation(
        b_points_uncorrected: pd.DataFrame,
        c_points: pd.DataFrame,
        statio_data: pd.DataFrame,
        outliers: pd.DataFrame,
        baseline: pd.DataFrame,
        sampling_rate_hz: float,
    ) -> pd.DataFrame:
        data = statio_data["statio_data"].to_frame()

        # insert NaN at the heartbeat id of the outliers
        data.loc[outliers.index, "statio_data"] = np.NaN

        # interpolate the outlier positions using linear interpolation
        data_interpol = (
            data["statio_data"]
            .astype(float)
            .interpolate(method="linear")
            .ffill()  # make sure that the first values are not NaN
            .bfill()  # make sure that the last values are not NaN
        )

        corrected_b_points = b_points_uncorrected.copy()

        # Add the baseline back to the interpolated values
        corrected_b_points.loc[data.index, "b_point_sample"] = (
            (c_points["c_point_sample"][c_points.loc[data.index].index] - (data_interpol + baseline) * sampling_rate_hz)
            .fillna(0)
            .astype(int)
        )
        corrected_b_points["b_point_sample"] = corrected_b_points["b_point_sample"].replace(0, np.nan)
        return corrected_b_points



"""Module for importing data recorded by the Biopac system."""

import datetime
from collections.abc import Sequence
from typing import ClassVar, Optional, Union

try:
    import bioread
except ImportError as e:
    raise ImportError(
        "The 'bioread' package is required to read Biopac data files. "
        "Please install it using 'pip install bioread' or 'poetry add bioread'."
    ) from e

import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_file_extension
from biopsykit.utils._types_internal import path_t, str_t

__all__ = ["BiopacDataset"]


class BiopacDataset:
    """Class for loading and processing Biopac data."""

    _CHANNEL_NAME_MAPPING: ClassVar[dict[str, str]] = {
        "ECG": "ecg",
        "RSP": "rsp",
        "EDA": "eda",
        "EMG": "emg",
        "ICG - Magnitude": "icg_mag",
        "ICG - Derivative": "icg_der",
        "SYNC": "sync",
    }

    _start_time_unix: pd.Timestamp
    _tz: str
    _event_markers: Optional[Sequence[bioread.reader.EventMarker]] = None
    _data: ClassVar[dict[str, pd.DataFrame]] = {}
    _sampling_rate: ClassVar[dict[str, int]] = {}

    def __init__(
        self,
        data_dict: dict[str, pd.DataFrame],
        sampling_rate_dict: dict[str, int],
        start_time: Optional[pd.Timestamp] = None,
        event_markers: Optional[Sequence[bioread.reader.EventMarker]] = None,
        tz: Optional[str] = None,
    ):
        """Get new Dataset instance.

        .. note::
            Usually you shouldn't use this init directly.
            Use the provided `from_acq_file` constructor to handle loading recorded Biopac Sessions.

        Parameters
        ----------
        data_dict : dict
            Dictionary containing data of the channels as :class:`pandas.DataFrame`.
            The keys of the dictionary are the channel names.
        sampling_rate_dict : dict
            Dictionary containing the sampling rate of the channels.
            The keys of the dictionary are the channel names.
        start_time : :class:`pandas.Timestamp`, optional
            Start time of the recording, if present, or ``None`` if no start time is available.
        event_markers : list of :class:`bioread.reader.EventMarker`, optional
            List of event markers set during the recording if present or ``None`` if no event markers are available.
        tz : str, optional
            Timezone of the recording, if present or ``None`` if no timezone is available.

        """
        self._data = data_dict
        for name, data in data_dict.items():
            setattr(self, name, data)
        for name, sampling_rate in sampling_rate_dict.items():
            setattr(self, f"sampling_rate_hz_{name}", sampling_rate)
        setattr(self, "channels", list(self._data.keys()))
        self._sampling_rate = sampling_rate_dict
        self._start_time_unix = start_time
        self._event_markers = event_markers
        self._tz = tz

    @classmethod
    def from_acq_file(
        cls, path: path_t, channel_mapping: Optional[dict[str, str]] = None, tz: Optional[str] = "Europe/Berlin"
    ):
        """Create a new Dataset from a valid .acq file.

        Parameters
        ----------
        path : :class:`pathlib.Path` or str
            Path to the file
        channel_mapping : dict, optional
            Dictionary containing the mapping of the channel names in the .acq to the channel names used in the Dataset.
        tz : str, optional
            Timezone str of the recording. This can be used to localize the start and end time.
            Note, this should not be the timezone of your current PC, but the timezone relevant for the specific
            recording.

        """
        # assert that file is an acq file
        _assert_file_extension(path, ".acq")
        biopac_data: bioread.reader.Datafile = bioread.read(str(path))

        start_time = None
        # if no event markers are available we can't compute a start time of the recording
        if biopac_data.event_markers is not None and len(biopac_data.event_markers) > 0:
            marker_time = pd.Timestamp(biopac_data.event_markers[0].date_created_utc)
            marker_sample_idx = biopac_data.event_markers[0].sample_index
            # start time is the marker time minus the time at the position of the marker sample
            start_time = marker_time - pd.Timedelta(seconds=biopac_data.time_index[marker_sample_idx])

        if channel_mapping is None:
            channel_mapping = cls._CHANNEL_NAME_MAPPING

        dict_channel_data, dict_sampling_rate = cls._extract_channel_information(biopac_data, channel_mapping)

        return cls(
            data_dict=dict_channel_data,
            sampling_rate_dict=dict_sampling_rate,
            start_time=start_time,
            event_markers=biopac_data.event_markers,
            tz=tz,
        )

    @property
    def start_time_unix(self) -> Optional[pd.Timestamp]:
        """Start time of the recording in UTC time."""
        return self._start_time_unix

    @property
    def timezone(self) -> str:
        """Timezone the dataset was recorded in."""
        return self._tz

    @property
    def event_markers(self):
        """Event markers set in the AcqKnowledge software during the recording."""
        return self._event_markers

    def data_as_df(
        self,
        datastreams: Optional[str_t] = None,
        index: Optional[str] = None,
        start_time: Optional[Union[str, datetime.datetime, pd.Timestamp]] = None,
    ) -> pd.DataFrame:
        """Return all data as one combined :class:`pandas.DataFrame`.

        Parameters
        ----------
        datastreams : str, optional
            name(s) of datastream to return in dataframe. If ``None``, all datastreams are returned.
        index : str, optional
            Specify which index should be used for the dataset. The options are:
            * "time": For the time in seconds since the first sample
            * "utc": For the utc time stamp of each sample
            * "utc_datetime": for a pandas DateTime index in UTC time
            * "local_datetime": for a pandas DateTime index in the timezone set for the session
            * None: For a simple index (0...N)
        start_time : str, :class:`datetime.datetime`, :class:`pandas.Timestamp`, optional
            Start time of the recording. Can be used to provide a custom start time if no start time can be inferred
            from the recording or to overwrite the start time extracted from the recording.

        """
        # sanitize datastreams input
        datastreams = self._sanitize_datastreams_input(datastreams)

        # assert that all datastreams have the same sampling rate
        sampling_rates = {self._sampling_rate[datastream] for datastream in datastreams}
        if len(sampling_rates) > 1:
            raise ValueError("All datastreams must have the same sampling rate for combining it into one DataFrame!")

        # get datastreams from dict
        data = [self._data[datastream] for datastream in datastreams]
        data = pd.concat(data, axis=1)

        data = self._add_index(data, index, start_time=start_time)
        return data

    @classmethod
    def _extract_channel_information(
        cls, biopac_data: bioread.reader.Datafile, channel_mapping: dict[str, str]
    ) -> tuple[dict[str, pd.DataFrame], dict[str, int]]:
        # TODO raise warning or error when there are more channels than were extracted
        #  (might be an indication that mapping does not contain all channels)
        dict_channel_data = {}
        dict_sampling_rate = {}
        for channel in biopac_data.channels:
            # check if channel name is in mapping
            for key, value in channel_mapping.items():
                if channel.name.startswith(key):
                    ch_name = value
                    channel_df = pd.DataFrame(
                        channel.data, index=pd.Index(biopac_data.time_index, name="t"), columns=[ch_name]
                    )
                    if ch_name in dict_channel_data:
                        if dict_sampling_rate[ch_name] != channel.samples_per_second:
                            raise ValueError(f"Sampling rates for '{ch_name}' must be the same for all channels!")
                        dict_channel_data[ch_name] = pd.concat([dict_channel_data[ch_name], channel_df], axis=1)
                    else:
                        dict_channel_data[ch_name] = channel_df

                    dict_sampling_rate[ch_name] = channel.samples_per_second
                    break

        return dict_channel_data, dict_sampling_rate

    def _add_index(self, data: pd.DataFrame, index: str, start_time: Optional[pd.Timestamp] = None) -> pd.DataFrame:
        index_names = {
            None: "n_samples",
            "time": "t",
            "utc": "utc",
            "utc_datetime": "date",
            "local_datetime": f"date ({self.timezone})",
        }
        if index and index not in index_names:
            raise ValueError(f"Supplied value for index ({index}) is not allowed. Allowed values: {index_names.keys()}")
        index_name = index_names[index]
        data.index.name = index_name

        if index == "time":
            return data
        if index is None:
            data = data.reset_index(drop=True)
            data.index.name = index_name
            return data

        if start_time is None:
            start_time = self.start_time_unix

        if start_time is None:
            raise ValueError(
                "No start time available - can't convert to datetime index! "
                "Use a different index representation or provide a custom start time using the 'start_time' parameter."
            )

        if index == "utc":
            # convert counter to utc timestamps
            data.index += start_time.timestamp()
            return data

        # convert counter to pandas datetime index
        data.index = pd.to_timedelta(data.index, unit="s")
        data.index += start_time

        if index == "local_datetime":
            data.index = data.index.tz_convert(self.timezone)

        return data

    def _sanitize_datastreams_input(self, datastreams) -> Sequence[str]:
        if datastreams is None:
            datastreams = list(self._data.keys())
        if isinstance(datastreams, str):
            # ensure list
            datastreams = [datastreams]
        # assert that all datastreams are available
        for datastream in datastreams:
            if datastream not in self._data:
                raise ValueError(f"Datastream '{datastream}' is not available in Dataset!")

        return datastreams



"""Module for sleep/wake detection base class."""

from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.dtypes import SleepWakeDataFrame


class _SleepWakeBase:
    """Base class for sleep/wake detection algorithms."""

    epoch_length: int = None
    """Epoch length in seconds."""

    def __init__(self, **kwargs):
        self.epoch_length: int = kwargs.get("epoch_length", 60)

    def fit(self, data: arr_t, **kwargs):
        """Fit sleep/wake detection algorithm to input data.

        .. note::
            Algorithms that do not have to (re)fit a ML model before sleep/wake prediction, such as rule-based
            algorithms, will internally bypass this method as the ``fit`` step is not needed.

        Parameters
        ----------
        data : array_like
            input data

        """
        raise NotImplementedError("Needs to be implemented by child class.")

    def predict(self, data: arr_t, **kwargs) -> SleepWakeDataFrame:
        """Apply sleep/wake prediction algorithm on input data.

        Parameters
        ----------
        data : array_like
            input data

        """
        raise NotImplementedError("Needs to be implemented by child class.")



import warnings

import numpy as np
import pandas as pd
from statsmodels.tools.sm_exceptions import ValueWarning
from statsmodels.tsa.ar_model import ar_select_order
from statsmodels.tsa.arima.model import ARIMA

from biopsykit.signals.icg.outlier_correction._base_outlier_correction import BaseOutlierCorrection

__all__ = ["OutlierCorrectionForouzanfar2018"]


from biopsykit.utils.dtypes import BPointDataFrame, CPointDataFrame, is_b_point_dataframe, is_c_point_dataframe

# TODO add verbosity option


class OutlierCorrectionForouzanfar2018(BaseOutlierCorrection):
    """B-Point outlier correction algorithm based on Forouzanfar et al. (2018) [1]_.

    This algorithm corrects outliers in B-Point data using an autoregressive model.

    References
    ----------
    .. [1] Forouzanfar, M., Baker, F. C., De Zambotti, M., McCall, C., Giovangrandi, L., & Kovacs, G. T. A. (2018).
        Toward a better noninvasive assessment of preejection period: A novel automatic algorithm for B-point detection
        and correction on thoracic impedance cardiogram. Psychophysiology, 55(8), e13072.
        https://doi.org/10.1111/psyp.13072

    """

    # @make_action_safe
    def correct_outlier(
        self,
        *,
        b_points: BPointDataFrame,
        c_points: CPointDataFrame,
        sampling_rate_hz: float,
        **kwargs,
    ):
        """Perform outlier correction on B-Point data.

        The outliers are corrected using an autoregressive model.

        The results of the outlier correction are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        b_points : :class:`~pandas.DataFrame`
            Extracted B-points. Each row contains the B-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. B-point locations can be NaN if no B-points were detected
            for certain heartbeats.
        c_points : :class:`~pandas.DataFrame`
            Extracted C-points. Each row contains the C-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. C-point locations can be NaN if no C-points were detected
            for certain heartbeats.
        sampling_rate_hz : int
            sampling rate of ICG derivative signal in hz
        kwargs : dict
            Additional keyword arguments:
                * verbose: bool, optional
                    Whether to print additional information. Default: False

        Returns
        -------
            self

        """
        verbose = kwargs.get("verbose", False)
        is_b_point_dataframe(b_points)
        is_c_point_dataframe(c_points)

        corrected_b_points = pd.DataFrame(index=b_points.index, columns=["b_point_sample"])

        # stationarize the B-Point time data
        stationary_data = self.stationarize_b_points(b_points, c_points, sampling_rate_hz)
        b_points_nan = b_points.loc[b_points["b_point_sample"].isna()]
        stationary_data.loc[b_points_nan.index, "statio_data"] = np.NaN

        # detect outliers
        outliers = self.detect_b_point_outlier(stationary_data)

        counter = 1
        if verbose:
            print(f"Detected {len(outliers)} outliers in correction cycle {counter}!")
        if len(outliers) == 0:
            is_b_point_dataframe(b_points)
            self.points_ = b_points
            return self

        # Perform the outlier correction until no more outliers are detected
        while len(outliers) > 0:
            if counter >= 30:
                break
            corrected_b_points = self._correct_outlier_autoregression(
                b_points, c_points, stationary_data, outliers, stationary_data["baseline"], sampling_rate_hz, **kwargs
            )
            stationary_data = self.stationarize_b_points(corrected_b_points, c_points, sampling_rate_hz)
            outliers = self.detect_b_point_outlier(stationary_data)
            if verbose:
                print(f"Detected {len(outliers)} outliers in correction cycle {counter}!")
            counter += 1

        if verbose:
            print("No more outliers got detected!")
        corrected_b_points = corrected_b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        is_b_point_dataframe(corrected_b_points)

        self.points_ = corrected_b_points
        return self

    @staticmethod
    def _correct_outlier_autoregression(
        b_points_uncorrected: pd.DataFrame,
        c_points: pd.DataFrame,
        statio_data: pd.DataFrame,
        outliers: pd.DataFrame,
        baseline: pd.DataFrame,
        sampling_rate_hz: int,
        **kwargs,
    ) -> pd.DataFrame:
        """Correct outliers in B-Point data using an autoregressive model.

        Parameters
        ----------
        b_points_uncorrected: :class:`~pandas.DataFrame`
            Dataframe containing the extracted B-Points per heartbeat, index functions as id of heartbeat
        c_points: :class:`~pandas.DataFrame`
            Dataframe containing the extracted C-Points per heartbeat, index functions as id of heartbeat
        statio_data: :class:`~pandas.DataFrame`
            Stationarized B-Point data
        outliers: :class:`~pandas.DataFrame`
            Dataframe containing the detected outliers
        baseline: :class:`~pandas.DataFrame`
            Baseline of the stationarized B-Point data
        sampling_rate_hz: int
            Sampling rate of ICG signal in hz
        **kwargs
            Additional keyword arguments:
                * verbose: bool, optional
                    Whether to print additional information. Default: False

        Returns
        -------
        :class:`~pandas.DataFrame`
            Corrected B-Point data

        """
        verbose = kwargs.get("verbose", False)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ValueWarning)
            warnings.simplefilter("ignore", category=FutureWarning)

            data = statio_data["statio_data"].to_frame()
            data.loc[outliers.index, "statio_data"] = np.NaN
            input_endog = (
                data["statio_data"]
                .astype(float)
                .interpolate(method="linear")
                .interpolate(method="ffill")  # ensure that no NaN values are at the beginning
                .interpolate(method="bfill")  # ensure that no NaN values are at the end
            )

            # Select order of the froward model
            maxlag = 30

            sel = None
            while maxlag > 0:
                try:
                    sel = ar_select_order(input_endog, maxlag=maxlag, ic="aic")
                except ValueError:
                    maxlag -= 1
                    if verbose:
                        print(f"Maxlag reduced to {maxlag}!")
                    continue
                break

            order = sel.ar_lags
            if order is None:
                order = [0]

            # fit the forward model
            arima_mod = ARIMA(endog=input_endog, order=(order, 0, 0))
            arima_res = arima_mod.fit(method="burg")
            # reverse the input data to get the backward model
            reversed_input = input_endog[::-1]
            # Select order of the backward model
            b_sel = ar_select_order(reversed_input, maxlag=maxlag, ic="aic")
            b_order = b_sel.ar_lags
            if b_order is None:
                b_order = [0]

            # Fit the backward model
            b_arima_mod = ARIMA(endog=reversed_input, order=(b_order, 0, 0))
            b_arima_res = b_arima_mod.fit(method="burg")

            # predict the outlier values
            for idx in outliers.index:
                forward_prediction = arima_res.predict(idx, idx)
                backward_idx = (len(reversed_input) - idx) % len(reversed_input)
                backward_prediction = b_arima_res.predict(backward_idx, backward_idx)
                prediction = np.average([forward_prediction, backward_prediction])
                data.loc[idx, "statio_data"] = prediction
            result = b_points_uncorrected.copy()
            result.loc[data.index, "b_point_sample"] = (
                (
                    c_points["c_point_sample"][c_points.loc[data.index].index]
                    - (data["statio_data"] + baseline) * sampling_rate_hz
                )
                .fillna(0)
                .astype(int)
            )
            result["b_point_sample"] = result["b_point_sample"].replace(0, np.nan)

            result = result.assign(maxlag=maxlag, padlen=statio_data["padlen"].iloc[0])

        return result



"""Module for processing EEG data."""

from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from biopsykit.signals._base import _BaseProcessor


class EegProcessor(_BaseProcessor):
    """Class for processing EEG data."""

    def __init__(
        self,
        data: Union[pd.DataFrame, dict[str, pd.DataFrame]],
        sampling_rate: Optional[float] = None,
        time_intervals: Optional[Union[pd.Series, dict[str, Sequence[str]]]] = None,
        include_start: Optional[bool] = False,
    ):
        """Initialize an ``EegProcessor`` instance.

        You can either pass a data dictionary 'data_dict' containing EEG data or dataframe containing
        EEG data. For the latter, you can additionally supply time information via ``time_intervals`` parameter
        to automatically split the data into single phases.


        Parameters
        ----------
        data : :class:`~pandas.DataFrame` or dict
            dataframe (or dict of such) with EEG data
        sampling_rate : float, optional
            sampling rate of recorded data
        time_intervals : dict or :class:`~pandas.Series`, optional
            time intervals indicating how ``data`` should be split.
            Can either be a :class:`~pandas.Series` with the `start` times of the single phases
            (the phase names are then derived from the index) or a dictionary with tuples indicating
            `start` and `end` times of phases (the phase names are then derived from the dict keys).
            Default: ``None`` (data is not split further)
        include_start : bool, optional
            ``True`` to include the data from the beginning of the recording to the first time interval as the
            first phase (then named ``Start``), ``False`` otherwise. Default: ``False``

        """
        super().__init__(
            data=data, sampling_rate=sampling_rate, time_intervals=time_intervals, include_start=include_start
        )

        self.eeg_result: dict[str, pd.DataFrame] = {}
        """Dictionary with EEG processing result dataframes, split into different phases.

        """

    def relative_band_energy(
        self,
        freq_bands: Optional[dict[str, Sequence[int]]] = None,
        title: Optional[str] = None,
    ) -> None:
        """Process EEG signal.

        Parameters
        ----------
        freq_bands : dict
            dictionary with frequency bounds of EEG frequency bands. By default (``None``) the following
            frequency band definition (in Hz) is used:

            * ``theta``: [4, 8]
            * ``alpha``: [8, 13]
            * ``beta``: [13, 30]
            * ``gamma``: [30, 44]

        title : str, optional
            title of ECG processing progress bar in Jupyter Notebooks or ``None`` to leave empty. Default: ``None``

        """
        from mne.time_frequency import psd_array_welch  # pylint:disable=import-outside-toplevel

        eeg_result = {}
        for key, df in tqdm(self.data.items(), desc=title):
            raw_array = np.transpose(df.values)
            # define the different frequency bands
            if freq_bands is None:
                freq_bands = {
                    # "delta": [1, 4],
                    "theta": [4, 8],
                    "alpha": [8, 13],
                    "beta": [13, 30],
                    "gamma": [30, 44],
                }

            # compute power spectral density using Welch's method
            psds, freqs = psd_array_welch(
                raw_array,
                sfreq=self.sampling_rate,
                fmin=0.5,
                fmax=self.sampling_rate / 2,
                average=None,
                n_overlap=int(0.9 * self.sampling_rate),
                verbose=0,
            )

            eeg_bands = []
            for fmin, fmax in freq_bands.values():
                # extract the FFT coefficients of the respective frequency band
                psds_band = psds[:, (freqs >= fmin) & (freqs < fmax)]
                # sum over the frequency bins
                psds_band = psds_band.sum(axis=1)
                # mean over the EEG channels
                psds_band = psds_band.mean(axis=0)
                eeg_bands.append(psds_band)

            eeg_bands = np.array(eeg_bands)

            # convert into dataframe
            df_bands = pd.DataFrame(np.transpose(eeg_bands), columns=list(freq_bands.keys()))
            # divide the band coefficients by the total sum of all frequency
            # band coefficients per sample to get the relative band powers
            df_bands = df_bands.div(df_bands.sum(axis=1), axis=0)

            # create a time axis and set as new index
            df_bands["timestamp"] = pd.to_datetime(
                np.linspace(
                    float(df.index[0].to_numpy()),
                    float(df.index[-1].to_numpy()),
                    len(df_bands),
                )
            )
            df_bands = df_bands.set_index("timestamp")
            df_bands = df_bands.tz_localize("UTC").tz_convert("Europe/Berlin")
            eeg_result[key] = df_bands

        self.eeg_result = eeg_result



"""I/O functions for files related to ECG processing."""

import re
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_is_dir
from biopsykit.utils._types_internal import path_t
from biopsykit.utils.dtypes import HeartRatePhaseDict, HeartRateSubjectDataDict, is_hr_phase_dict
from biopsykit.utils.file_handling import get_subject_dirs, is_excel_file
from biopsykit.utils.time import tz

__all__ = [
    "load_hr_phase_dict",
    "load_hr_phase_dict_csv",
    "load_hr_phase_dict_folder",
    "write_hr_phase_dict",
    "write_hr_phase_dict_csv",
]


def load_hr_phase_dict(file_path: path_t, assert_format: Optional[bool] = True) -> HeartRatePhaseDict:
    """Load Excel file containing time series heart rate data of one subject.

    The returned dictionary will be a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`,
    i.e., a dict with heart rate data from one subject split into phases (as exported by :func:`write_hr_phase_dict`
    or :func:`write_hr_phase_dict_csv`).

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to file
    assert_format : bool, optional
        whether to check if the imported dict is in the right format or not

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        Dict with heart rate data split into phases

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if file in ``file_path`` is not a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        (if ``assert_format`` is ``True``)
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no Excel file (`.xls` or `.xlsx`)

    See Also
    --------
    ~biopsykit.utils.dtypes.HeartRatePhaseDict : Dictionary format
    write_hr_phase_dict : Write :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` to file

    """
    # ensure pathlib
    file_path = Path(file_path)
    _assert_file_extension(file_path, (".xls", ".xlsx"))

    # assure that the file is an Excel file
    is_excel_file(file_path)

    dict_hr: HeartRatePhaseDict = pd.read_excel(file_path, index_col="time", sheet_name=None)
    if assert_format:
        # assert that the dictionary is in the correct format
        is_hr_phase_dict(dict_hr)

    # (re-)localize each sheet since Excel does not support timezone-aware dates
    dict_hr = {k: v.tz_localize(tz) for k, v in dict_hr.items()}
    return dict_hr


def load_hr_phase_dict_folder(
    base_path: path_t, filename_pattern: str, subfolder_pattern: Optional[str] = None
) -> HeartRateSubjectDataDict:
    r"""Load a folder with multiple ``HeartRatePhaseDict`` and concatenate them  into a ``HeartRateSubjectDataDict``.

    This functions looks for all files that match the ``file_pattern`` in the folder specified by ``base_path``
    and loads the files that are all expected to be :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`.

    Subject IDs are extracted from the file name. Hence, ``file_pattern`` needs to be a regex
    including a capture group, e.g. "ecg_results_(\\w+).xlsx".

    Alternatively, if the files are stored in subfolders, the name pattern of these subfolders can be specified by
    ``subject_folder_pattern``. Then, it is expected that the subfolder names correspond to the subject IDs.

    The returned dictionary will be a :obj:`~biopsykit.utils.dtypes.HeartRateSubjectDataDict`
    with the following format:

    { ``subject_id`` : :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` }


    Parameters
    ----------
    base_path : :class: `~pathlib.Path` or str
        path to top-level folder containing all subject folders
    filename_pattern : str
        filename pattern of exported :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`.
        Must be a regex string with capture group to extract subject IDs, or a regular regex string
        (without capture group) if ``subfolder_pattern`` is specified
    subfolder_pattern : str, optional
        subfolder name pattern if files are stored in subfolders.
        Then, ``filename_pattern`` does **not** need to be a regex with a capture group because it is assumed that
        the names of the subfolders correspond to the subject IDs.

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.HeartRateSubjectDataDict`
        :obj:`~biopsykit.utils.dtypes.HeartRateSubjectDataDict`, i.e., a dictionary with
        :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` of multiple subjects

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if any file that matches ``filename_pattern`` is not a
        :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
    :exc:`FileNotFoundError`
        if no files match ``filename_pattern`` or no subfolders match ``subfolder_pattern``

    See Also
    --------
    ~biopsykit.utils.file_handling.get_subject_dirs : Filter for subject subfolders in a given folder
    load_hr_phase_dict : Load :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` for one subject

    Examples
    --------
    >>> from biopsykit.io.ecg import load_hr_phase_dict_folder
    >>> base_path = "./ecg_results/"
    >>> # Option 1: all files are stored in `base_path`, subject IDs are extracted from the file names
    >>> dict_hr_subjects = load_hr_phase_dict_folder(
    >>>                         base_path,
    >>>                         filename_pattern=r"ecg_result_(\\w+).xlsx")
    >>> print(dict_hr_subjects)
    {
         'Vp01': {}, # one single HeartRatePhaseDict
         'Vp02': {},
         # ...
    }
    >>> # Option 2: files are stored in subfolders, the name of the subfolders is the corresponding subject ID
    >>> dict_hr_subjects = load_hr_phase_dict_folder(
    >>>                         base_path,
    >>>                         filename_pattern=r"ecg_result*.xlsx",
    >>>                         subfolder_pattern="Vp*")
    >>> print(dict_hr_subjects)
    {
         'Vp01': {}, # one single HeartRatePhaseDict
         'Vp02': {},
         # ...
    }

    """
    # ensure pathlib
    base_path = Path(base_path)

    dict_hr_subjects = {}
    if subfolder_pattern is None:
        file_list = sorted(base_path.glob("*"))
        file_list = [f for f in file_list if re.search(filename_pattern, f.name)]
        if len(file_list) == 0:
            raise FileNotFoundError(f"No files matching the pattern '{filename_pattern}' found in {base_path}.")
        for file in file_list:
            subject_id = re.findall(filename_pattern, file.name)[0]
            dict_hr_subjects[subject_id] = load_hr_phase_dict(file)
    else:
        subject_dirs = get_subject_dirs(base_path, subfolder_pattern)

        for subject_dir in subject_dirs:
            subject_id = subject_dir.name
            hr_phase_dict = _load_hr_phase_dict_single_subject(subject_dir, filename_pattern)
            if hr_phase_dict is None:
                continue
            dict_hr_subjects[subject_id] = hr_phase_dict
    return dict_hr_subjects


def _load_hr_phase_dict_single_subject(subject_dir: Path, filename_pattern: str) -> Optional[HeartRatePhaseDict]:
    subject_id = subject_dir.name
    # first try to search for files with glob (assuming that a regex string without capture group was passed),
    # then try to search via regex search (assuming that a regex string with capture group was passed,
    # which should actually not be done if subfolder_pattern is passed)
    file_list = sorted(subject_dir.glob(filename_pattern))
    if len(file_list) == 0:
        file_list = sorted(subject_dir.glob("*"))
        # then extract the ones that match
        file_list = [f for f in file_list if re.search(filename_pattern, f.name)]

    if len(file_list) == 1:
        return load_hr_phase_dict(file_list[0])
    if len(file_list) > 1:
        warnings.warn(
            f'More than one file matching file pattern "{filename_pattern}" found in folder {subject_dir}. '
            "Trying to merge these files into one HeartRatePhaseDict"
        )
        dict_hr = {}
        for file in file_list:
            dict_hr.update(load_hr_phase_dict(file))
        return dict_hr

    print(f"No Heart Rate data for subject {subject_id}")
    return None


def write_hr_phase_dict(hr_phase_dict: HeartRatePhaseDict, file_path: path_t):
    """Write :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` to an Excel file.

    The :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` is a dictionary with heart rate time
    series data split into phases.

    Each of the phases in the dictionary will be a separate sheet in the Excel file.

    Parameters
    ----------
    hr_phase_dict : :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` containing pandas dataframes
        with heart rate data
    file_path : :class:`~pathlib.Path` or str
        path to export file

    See Also
    --------
    ~biopsykit.utils.dtypes.HeartRatePhaseDict : Dictionary format
    load_hr_phase_dict : Load :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` written to file
    ~biopsykit.io.write_pandas_dict_excel : Write dictionary with pandas dataframes to Excel file

    """
    from biopsykit.io import write_pandas_dict_excel  # pylint:disable=import-outside-toplevel

    # assert that file path is an Excel file
    is_excel_file(file_path)

    # assert that dict is in the right format
    is_hr_phase_dict(hr_phase_dict)
    write_pandas_dict_excel(hr_phase_dict, file_path)


def write_hr_phase_dict_csv(hr_phase_dict: HeartRatePhaseDict, folder_path: path_t, file_pattern: path_t):
    """Write :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` to a series of csv files.

    The :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` is a dictionary with heart rate time
    series data split into phases.

    Each of the phases in the dictionary will be a separate csv file.

    Parameters
    ----------
    hr_phase_dict : :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` containing pandas dataframes
        with heart rate data
    folder_path : :class:`~pathlib.Path` or str
        folder path to export csv files
    file_pattern : :class:`~pathlib.Path` or str
        file pattern to save. The file pattern must include a placeholder "{}" which will be filled with the
        name of the phase.

    Raises
    ------
    ValueError
        if ``file_pattern`` does not include a placeholder "{}" that can be filled with the phase name
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if ``hr_phase_dict`` is not a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if ``file_pattern`` is no csv file

    See Also
    --------
    ~biopsykit.utils.dtypes.HeartRatePhaseDict : Dictionary format
    load_hr_phase_dict_csv : Load :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` written to csv files

    """
    # ensure pathlib
    folder_path = Path(folder_path)
    _assert_is_dir(folder_path)

    # check file pattern for sanity
    file_pattern = str(file_pattern)
    if "{}" not in file_pattern:
        raise ValueError(
            f"'file_pattern' must include a placeholder to be filled with the phase name!. Got {file_pattern}."
        )
    is_hr_phase_dict(hr_phase_dict)

    for phase, data in hr_phase_dict.items():
        file_name = Path(file_pattern.format(phase))
        file_path = folder_path.joinpath(file_name)
        _assert_file_extension(file_path, ".csv")
        data.to_csv(file_path)


def load_hr_phase_dict_csv(
    folder_path: path_t,
    file_pattern: path_t,
    phase_order: Optional[Sequence[str]] = None,
    assert_format: Optional[bool] = True,
) -> HeartRatePhaseDict:
    """Load csv file with time series HR data of one subject from folder and combine it into a ``HeartRatePhaseDict``.

    The returned dictionary will be a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`,
    i.e., a dict with heart rate data from one subject split into phases (as exported by :func:`write_hr_phase_dict`).

    Parameters
    ----------
    folder_path : :class:`~pathlib.Path` or str
        folder path to export csv files
    file_pattern : :class:`~pathlib.Path` or str
        file pattern of the csv files. `file_pattern` must include a regex capture group (see Examples) which is used
        to extract the phase name from the file name.
    phase_order : list of str, optional
        list of phase names to order resulting :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` or ``None``
        to order phases according to the file name ordering in ``folder_path``.
        Default: ``None``
    assert_format : bool, optional
        whether to check if the imported dict is in the right format or not

    Returns
    -------
    :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        Dict with heart rate data split into phases

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.ValidationError`
        if file in ``file_path`` is not a :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict`
        (if ``assert_format`` is ``True``)
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if file is no Excel file (`.xls` or `.xlsx`)

    See Also
    --------
    ~biopsykit.utils.dtypes.HeartRatePhaseDict : Dictionary format
    write_hr_phase_dict_csv : Write :obj:`~biopsykit.utils.dtypes.HeartRatePhaseDict` to a series of csv files

    """
    # ensure pathlib
    folder_path = Path(folder_path)
    _assert_is_dir(folder_path)

    file_list = sorted(folder_path.glob("*.csv"))
    if len(file_list) == 0:
        raise ValueError(f"No csv files found in {folder_path}!")

    dict_hr = {}
    for file in file_list:
        phase_name = re.findall(file_pattern, file.name)
        if len(phase_name) == 0:
            raise ValueError(f"No string matching the pattern {file_pattern} found in {file}!")
        if len(phase_name) > 1:
            raise ValueError(f"Too many strings matching the pattern {file_pattern} found in {file}!")
        phase_name = phase_name[0]
        df = pd.read_csv(file, index_col="time")
        df.index = pd.to_datetime(df.index, errors="ignore")
        dict_hr[phase_name] = df

    if assert_format:
        # assert that the dictionary is in the correct format
        is_hr_phase_dict(dict_hr)

    if phase_order is not None:
        _assert_phase_names(folder_path, phase_order, dict_hr.keys())
        dict_hr = {phase: dict_hr[phase] for phase in phase_order}

    return dict_hr


def _assert_phase_names(folder_path, phase_order, dict_keys):
    if len(phase_order) != len(dict_keys):
        raise ValueError(
            f"Number of phases provided by 'phase_order' do not match. "
            f"Expected {len(dict_keys)}, got {len(phase_order)}."
        )
    if set(phase_order) != set(dict_keys):
        raise ValueError(
            f"Phases provided by 'phase_order' do not match the phases of the files in {folder_path}. "
            f"Expected {phase_order}, got {list(dict_keys)}."
        )



"""Helper functions for file handling."""

import re
from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

import matplotlib.pyplot as plt

from biopsykit.utils._datatype_validation_helper import _assert_file_extension
from biopsykit.utils._types_internal import path_t


def mkdirs(dir_list: Union[path_t, Sequence[path_t]]) -> None:
    """Batch-create a list of directories.

    Conveniently create a list of directories, e.g. directories for storing processing
    results, figures, and statistic reports, using :meth:`~pathlib.Path.mkdir`.
    If parent directories do not exist yet, they are created (`pathlib` option ``parent=True``),
    if directories already exist no Error is thrown (`pathlib` option ``exist_ok=True``).

    Parameters
    ----------
    dir_list : list of path or str
        list of directory names to create

    Examples
    --------
    >>> from biopsykit.utils.file_handling import mkdirs
    >>> path_list = [Path("processing_results"), Path("exports/figures"), Path("exports/statistics")]
    >>> mkdirs(path_list)

    """
    if isinstance(dir_list, (str, Path)):
        dir_list = [dir_list]
    for directory in dir_list:
        # ensure pathlib
        directory_path = Path(directory)
        directory_path.mkdir(exist_ok=True, parents=True)


def get_subject_dirs(base_path: path_t, pattern: str) -> Optional[Sequence[Path]]:
    """Filter for subject directories using a name pattern.

    Parameters
    ----------
    base_path : path or str
        base path to filter for directories
    pattern : str
        name pattern as regex

    Returns
    -------
    list of path
        a list of path or an empty list if no subfolders matched the ``pattern``

    Raises
    ------
    FileNotFoundError
        if no subfolders in ``base_path`` match ``pattern``.

    Examples
    --------
    >>> from biopsykit.utils.file_handling import get_subject_dirs
    >>> base_path = Path(".")
    >>> get_subject_dirs(base_path, "Vp*")

    """
    # ensure pathlib
    base_path = Path(base_path)
    subject_dirs = [p for p in sorted(base_path.glob("*")) if p.is_dir()]
    subject_dirs = list(filter(lambda s: len(re.findall(pattern, s.name)) > 0, subject_dirs))
    if len(subject_dirs) == 0:
        raise FileNotFoundError(f"No subfolders matching the pattern '{pattern}' found in {base_path}.")
    return subject_dirs


def export_figure(
    fig: plt.Figure,
    filename: path_t,
    base_dir: path_t,
    formats: Optional[Sequence[str]] = None,
    use_subfolder: Optional[bool] = True,
    **kwargs,
):
    """Export matplotlib figure to file(s).

    This function allows to export a matplotlib figure in multiple file formats, each format being optionally saved
    in its own output subfolder.

    Parameters
    ----------
    fig : :class:`~matplotlib.figure.Figure`
        matplotlib figure object
    filename : :class:`~pathlib.Path` or str
        name of the output file
    base_dir : path or str
        base directory to export figures to
    formats: list of str, optional
        list of file formats to export or ``None`` to export as pdf. Default: ``None``
    use_subfolder : bool, optional
        whether to create an own output subfolder per file format or not. Default: ``True``
    **kwargs
        additional keyword arguments to pass to :meth:`~matplotlib.figure.Figure.savefig`

    Examples
    --------
    >>> fig = plt.Figure()
    >>>
    >>> base_dir = "./img"
    >>> filename = "plot"
    >>> formats = ["pdf", "png"]
    >>> # Export into subfolders (default)
    >>> export_figure(fig, filename=filename, base_dir=base_dir, formats=formats)
    >>> # | img/
    >>> # | - pdf/
    >>> # | - - plot.pdf
    >>> # | - png/
    >>> # | - - plot.png

    >>> # Export into one folder
    >>> export_figure(fig, filename=filename, base_dir=base_dir, formats=formats, use_subfolder=False)
    >>> # | img/
    >>> # | - plot.pdf
    >>> # | - plot.png

    """
    if formats is None:
        formats = ["pdf"]
    # ensure list
    if isinstance(formats, str):
        formats = [formats]

    # ensure pathlib
    base_dir = Path(base_dir)
    filename = Path(filename)
    subfolders = [base_dir] * len(formats)

    if use_subfolder:
        subfolders = [base_dir.joinpath(f) for f in formats]
        for folder in subfolders:
            folder.mkdir(exist_ok=True, parents=True)

    for f, subfolder in zip(formats, subfolders):
        fig.savefig(
            subfolder.joinpath(filename.name + "." + f),
            transparent=(f == "pdf"),
            format=f,
            **kwargs,
        )


def is_excel_file(file_name: path_t, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether the file name is an Excel file.

    Parameters
    ----------
    file_name : :class:`~pathlib.Path` or str
        file name to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``file_name`` is an Excel file, i.e. has the suffix ``.xlsx``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``file_name`` is not an Excel file

    """
    return _assert_file_extension(file_name, (".xlsx", ".xls"), raise_exception)



"""Module for loading and processing Task Force Monitor (TFM) data."""

from typing import ClassVar, Optional

import pandas as pd
from scipy.io import loadmat

from biopsykit.utils._types_internal import path_t


class TFMDataset:
    """Class for loading and processing Task Force Monitor (TFM) data."""

    CHANNEL_MAPPING: ClassVar[dict[str, str]] = {"ecg_1": "rawECG1", "ecg_2": "rawECG2", "icg_der": "rawICG"}
    _tz: str

    def __init__(
        self, data_dict: dict[str, pd.DataFrame], sampling_rate_dict: dict[str, float], tz: Optional[str] = None
    ):
        """Initialize a TFM dataset.

        Parameters
        ----------
        data_dict : dict
            Dictionary containing TFM data. Keys are channel names, values are dataframes with the TFM data.
        tz : str, optional
            Timezone of the data. Default: None

        """
        self._data = data_dict
        self._sampling_rate = sampling_rate_dict
        for name, data in data_dict.items():
            setattr(self, name, data)
        for name, sampling_rate in sampling_rate_dict.items():
            setattr(self, f"sampling_rate_hz_{name}", sampling_rate)
        setattr(self, "channels", list(self._data.keys()))

        self._tz = tz

    @classmethod
    def from_mat_file(
        cls,
        path: path_t,
        # channel_mapping: Optional[Dict[str, str]] = None,
        tz: Optional[str] = "Europe/Berlin",
    ):
        """Load a TFM dataset from a .mat file.

        Parameters
        ----------
        path : str or :class:`~pathlib.Path`
            Path to the .mat file.
        tz : str, optional
            Timezone of the data. Default: "Europe/Berlin"

        """
        data = loadmat(path, struct_as_record=False, squeeze_me=True)
        data_raw = data["RAW_SIGNALS"]
        # keys = [s for s in dir(data_raw) if not s.startswith("_")]
        # print(keys)
        data_dict = {key: getattr(data_raw, value) for key, value in cls.CHANNEL_MAPPING.items()}
        return cls(data_dict=data_dict, tz=tz, sampling_rate_dict={})

    def data_as_df(self) -> dict[str, pd.DataFrame]:
        """Return the TFM data as a dictionary of pandas DataFrames.

        Returns
        -------
        dict
            Dictionary containing the TFM data as pandas DataFrames. Keys are channel names, values are the dataframes.

        """
        return self._data



"""Module for importing data recorded by a PSG system (expects .edf files)."""

import time
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

try:
    import mne
except ImportError as e:
    raise ImportError(
        "The 'mne' package is required to read edf data files. "
        "Please install it using 'pip install mne' or 'poetry add mne'."
    ) from e

import pandas as pd

from biopsykit.utils._datatype_validation_helper import _assert_file_extension, _assert_is_dir
from biopsykit.utils._types_internal import path_t

__all__ = ["PSGDataset"]


class PSGDataset:
    """Class for loading and processing PSG data."""

    def __init__(
        self,
        data_dict: dict[str, pd.DataFrame],
        sampling_rate_dict: dict[str, int],
        start_time: Optional[pd.Timestamp] = None,
        tz: Optional[str] = "Europe/Berlin",
    ):
        self._data = data_dict
        for name, data in data_dict.items():
            setattr(self, name, data)
        for _, sampling_rate in sampling_rate_dict.items():
            setattr(self, "sampling_rate", sampling_rate)
        setattr(self, "channels", list(self._data.keys()))
        self._sampling_rate = sampling_rate_dict
        self._start_time_datetime = pd.Timestamp(start_time)
        self._start_time_unix = int(time.mktime(start_time.timetuple()))
        self._tz = tz

    @classmethod
    def from_edf_file(
        cls,
        path: path_t,
        datastreams: Optional[Sequence] = None,
        tz: Optional[str] = "Europe/Berlin",
    ):
        """Create a new Dataset from a valid .edf file.

        Parameters
        ----------
        path : :class:`pathlib.Path` or str
            Path to the .edf file
        datastreams : Optional[Sequence], optional
            List of datastreams to load, by default None. If None, all datastreams are loaded.
        tz : Optional[str], optional
            Timezone of the recording, by default "Europe/Berlin".

        Returns
        -------
        PSGDataset : New Dataset instance

        """
        if path.is_dir():
            data_dict, fs, start_time = cls.load_data_folder(path, datastreams)
        else:
            data_dict, fs, start_time = cls.load_data(path, datastreams)

        return cls(data_dict=data_dict, sampling_rate_dict={"sampling_rate": fs}, start_time=start_time, tz=tz)

    @property
    def start_time_unix(self) -> Optional[pd.Timestamp]:
        """Start time of the recording in UTC time."""
        return self._start_time_unix

    @property
    def start_time_datetime(self) -> Optional[pd.Timestamp]:
        """Start time of the recording in UTC time."""
        return self._start_time_datetime

    @property
    def timezone(self) -> str:
        """Timezone the dataset was recorded in."""
        return self._tz

    def data_as_df(self, index: Optional[str] = None) -> pd.DataFrame:
        """Return data as one combined pandas.DataFrame.

        Parameters
        ----------
        index : str, optional
            Specify which index should be used for the dataset. The options are:
            * "time": For the time in seconds since the first sample
            * "utc": For the utc time stamp of each sample
            * "utc_datetime": for a pandas DateTime index in UTC time
            * "local_datetime": for a pandas DateTime index in the timezone set for the session
            * None: For a simple index (0...N)

        Returns
        -------
        pd.DataFrame: Combined data as pandas DataFrame with respective index

        """
        # get datastreams from dict
        datastreams = self._data.keys()
        data = [self._data[datastream] for datastream in datastreams]
        data = pd.concat(data, axis=1)

        data = self._add_index(data, index)

        return data

    def load_ground_truth(
        self,
        path: path_t,
    ):
        """Load ground truth data from a .xlsx file which can be exported from the Software Somnomedics.

        .. note::
            Other formats are not supported yet and raise a FileNotFoundError.

        path : :class:`pathlib.Path` or str
            path to the .xlsx file
        return: pd.DataFrame
            ground truth data as pandas DataFrame

        """
        file_path = path.parents[1].joinpath("labels/PSG_analyse.xlsx")
        try:
            sleep_phases = pd.read_excel(file_path, sheet_name="Schlafprofil", header=7, index_col=0)
        except FileNotFoundError:
            warnings.warn("No ground truth found")
            return pd.DataFrame()

        return sleep_phases
        # TODO: Read in excel or txt files with sleep labels

    @classmethod
    def load_data_folder(
        cls,
        folder_path: path_t,
        datastreams: Optional[Sequence] = None,
    ):
        """Load data from a folder containing a single .edf file.

        Parameters
        ----------
        folder_path : :class:`pathlib.Path` or str
            path to the folder containing the .edf file
        datastreams: lst, optional
            list of datastreams to load. If None, all datastreams are loaded
        timezone: str, optional
            timezone of the recording. If None, the timezone is set to UTC

        Returns
        -------
        data_dict: dict of datastreams
        fs: sampling rate
        start_time: start time of the recording

        Raises
        ------
        FileNotFoundError: if no .edf file is found in the folder
        ValueError: if more than one .edf file is found in the folder

        """
        _assert_is_dir(folder_path)

        # look for all PSG .edf files in the folder
        dataset_list = sorted(folder_path.glob("*.edf"))
        if len(dataset_list) == 0:
            raise FileNotFoundError(f"No PSG files found in folder {folder_path}!")
        if len(dataset_list) > 1:
            raise ValueError(
                f"More than one PSG files found in folder {folder_path}!"
                f"This function only supports one recording per folder!"
            )
        result_dict, fs, start_time = cls.load_data(folder_path.joinpath(dataset_list[0]), datastreams)

        return result_dict, fs, start_time

    @classmethod
    def load_data(
        cls,
        path: path_t,
        datastreams: Optional[Sequence] = None,
    ):
        """Load PSG data from a valid .edf file.

        Parameters
        ----------
        path : :class:`pathlib.Path` or str
            path to the .edf file
        datastreams: lst, optional
            list of datastreams to load. If None, all datastreams are loaded
        timezone: str, optional
            timezone of the recording. If None, the timezone is set to UTC

        Returns
        -------
        data_dict: dict of datastreams
        fs: sampling rate
        start_time: start time of the recording

        Raises
        ------
        Value Error: Not all datastreams are found in the .edf file

        """
        # load raw data
        data_psg, fs = cls.load_data_raw(path)

        # select datastreams to extract
        if datastreams is None:
            datastreams = data_psg.ch_names
        if isinstance(datastreams, str):
            datastreams = [datastreams]

        # save extracted datastreams in dict
        result_dict = {}
        for datastream in datastreams:
            try:
                time_idx, _, start_time = cls._create_dt_index(data_psg.info["meas_date"], times_array=data_psg.times)
                psg_datastream = data_psg.copy().pick([datastream]).get_data()[0, :]
                result_dict[datastream] = pd.DataFrame(psg_datastream, index=time_idx, columns=[datastream])
            except ValueError:
                print(
                    "Not all channels match the selected datastreams - Following Datastreams are available: "
                    + str(data_psg.ch_names)
                )

        return result_dict, fs, start_time

    @classmethod
    def load_data_raw(
        cls,
        path: path_t,
    ):
        """Load PSG data from .edf file.

        Parameters
        ----------
        path : :class:`pathlib.Path` or str
            path to the .edf file
        timezone: str, optional
            timezone of the recording. If None, the timezone is set to UTC

        Returns
        -------
        data: mne.io.Raw object
        fs: sampling rate

        """
        # ensure pathlib
        path = Path(path)
        _assert_file_extension(path, ".edf")

        # load data from edf file
        edf = mne.io.read_raw_edf(path)

        # get sampling rate
        fs = edf.info["sfreq"]

        return edf, fs

    @classmethod
    def _create_dt_index(cls, starttime, times_array):
        """Create a datetime index from the start time and the times array."""
        starttime_s = starttime.timestamp()
        # add start time to array of timestamps
        times_array = times_array + starttime_s
        # convert to datetime
        datetime_index = pd.to_datetime(times_array, unit="s")
        # generate epochs from datetime index
        epochs, start_time = cls._generate_epochs(datetime_index)
        return datetime_index, epochs, start_time

    @classmethod
    def _generate_epochs(cls, datetime_index):
        """Generate epochs from a datetime index."""
        start_time = datetime_index[0]
        # round to 30 second epochs
        epochs_30s = datetime_index.round("30s")

        epochs_clear = (epochs_30s - start_time).total_seconds()

        epochs_clear = epochs_clear / 30
        epochs = epochs_clear.astype(int)
        return epochs, start_time

    def _add_index(self, data: pd.DataFrame, index: str) -> pd.DataFrame:
        index_names = {
            None: "n_samples",
            "time": "t",
            "utc": "utc",
            "utc_datetime": "date",
            "local_datetime": f"date ({self.timezone})",
        }
        if index and index not in index_names:
            raise ValueError(f"Supplied value for index ({index}) is not allowed. Allowed values: {index_names.keys()}")
        index_name = index_names[index]
        data.index.name = index_name
        if index is None:
            data = data.reset_index(drop=True)
            data.index.name = index_name
            return data
        if index == "utc_datetime":
            data.index = data.index.tz_localize(self.timezone).tz_convert("UTC")
            return data
        if index == "time":
            data.index = data.index - self.start_time_datetime
            data.index = data.index.total_seconds()
            return data
        if index == "utc":
            # convert counter to utc timestamps i seconds
            data = data.reset_index(drop=True)
            data.index.astype("int64")
            data.index += self.start_time_unix
            return data
        if index == "local_datetime":
            data.index = data.index.tz_localize(self.timezone)
            return data
        return data



"""Module for processing Respiration data."""

from collections.abc import Sequence
from typing import Optional, Union

import neurokit2 as nk
import numpy as np
import pandas as pd
import scipy.signal as ss

from biopsykit.signals._base import _BaseProcessor
from biopsykit.utils.array_handling import sanitize_input_1d

__all__ = ["RspProcessor"]


class RspProcessor(_BaseProcessor):
    """Class for processing Respiration data."""

    def __init__(
        self,
        data: Union[pd.DataFrame, dict[str, pd.DataFrame]],
        sampling_rate: Optional[float] = None,
        time_intervals: Optional[Union[pd.Series, dict[str, Sequence[str]]]] = None,
        include_start: Optional[bool] = False,
    ):
        """Initialize an ``RspProcessor`` instance.

        You can either pass a data dictionary 'data_dict' containing Respiration data or dataframe containing
        Respiration data. For the latter, you can additionally supply time information via ``time_intervals`` parameter
        to automatically split the data into single phases.


        Parameters
        ----------
        data : :class:`~pandas.DataFrame` or dict
            dataframe (or dict of such) with Respiration data
        sampling_rate : float, optional
            sampling rate of recorded data
        time_intervals : dict or :class:`~pandas.Series`, optional
            time intervals indicating how ``data`` should be split.
            Can either be a :class:`~pandas.Series` with the `start` times of the single phases
            (the phase names are then derived from the index) or a dictionary with tuples indicating
            `start` and `end` times of phases (the phase names are then derived from the dict keys).
            Default: ``None`` (data is not split further)
        include_start : bool, optional
            ``True`` to include the data from the beginning of the recording to the first time interval as the
            first phase (then named ``Start``), ``False`` otherwise. Default: ``False``

        """
        super().__init__(
            data=data, sampling_rate=sampling_rate, time_intervals=time_intervals, include_start=include_start
        )

    @classmethod
    def rsp_compute_rate(cls, rsp_signal: pd.DataFrame, sampling_rate: Optional[float] = 256.0) -> float:
        """Compute respiration rate for given interval from respiration signal.

        Parameters
        ----------
        rsp_signal : :class:`~pandas.DataFrame`
            Raw respiration signal (1D). Can be a 'true' respiration signal (e.g. from bioimpedance or Radar)
            or an 'estimated' respiration signal (e.g. from ECG-derived respiration)
        sampling_rate : float, optional
            Sampling rate of recorded data


        Returns
        -------
        float
            Respiration rate during the given interval in bpm (breaths per minute)


        References
        ----------
        Schäfer, A., & Kratky, K. W. (2008). Estimation of Breathing Rate from Respiratory Sinus Arrhythmia:
        Comparison of Various Methods. *Annals of Biomedical Engineering*, 36(3), 476-485.
        https://doi.org/10.1007/s10439-007-9428-1


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> # Extract respiration signal estimated from ECG using the 'peak_trough_diff' method
        >>> rsp_signal = ecg_processor.ecg_estimate_rsp(ecg_processor, key="Data", edr_type='peak_trough_diff')
        >>> # Compute respiration rate from respiration signal
        >>> rsp_rate = ecg_processor.rsp_compute_rate(rsp_signal)

        """
        # find peaks: minimal distance between peaks: 1 seconds
        rsp_signal = sanitize_input_1d(rsp_signal)
        edr_maxima = ss.find_peaks(rsp_signal, height=0, distance=sampling_rate)[0]
        edr_minima = ss.find_peaks(-1 * rsp_signal, height=0, distance=sampling_rate)[0]
        # threshold: 0.2 * Q3 (= 75th percentile)
        max_threshold = 0.2 * np.percentile(rsp_signal[edr_maxima], 75)
        # find all maxima that are above the threshold
        edr_maxima = edr_maxima[rsp_signal[edr_maxima] > max_threshold]

        # rearrange maxima into 2D array to that each row contains the start and end of one cycle
        rsp_cycles_start_end = np.vstack([edr_maxima[:-1], edr_maxima[1:]]).T

        # check for
        valid_resp_phases_mask = np.apply_along_axis(cls._check_contains_trough, 1, rsp_cycles_start_end, edr_minima)
        rsp_cycles_start_end = rsp_cycles_start_end[valid_resp_phases_mask]

        edr_signal_split: list = np.split(rsp_signal, rsp_cycles_start_end.flatten())[1::2]
        rsp_cycles_start_end = rsp_cycles_start_end.T
        rsp_peaks = rsp_cycles_start_end[0]
        rsp_troughs = np.array(list(map(np.argmin, edr_signal_split))) + rsp_cycles_start_end[0]

        rsp_rate_peaks = cls._rsp_rate(rsp_peaks, int(sampling_rate), len(rsp_signal))
        rsp_rate_troughs = cls._rsp_rate(rsp_troughs, int(sampling_rate), len(rsp_signal))
        return np.concatenate([rsp_rate_peaks, rsp_rate_troughs]).mean()

    @classmethod
    def _check_contains_trough(cls, start_end: np.ndarray, minima: np.ndarray) -> bool:
        """Check whether exactly one minima is in the interval given by the array ``start_end``: [start, end].

        Parameters
        ----------
        start_end : :class:`numpy.array`
            Array with start and end index
        minima : :class:`numpy.array`
            Array containing minima to be checked


        Returns
        -------
        bool
            ``True`` if exactly one minima is in the ``[start, end]`` interval, ``False`` otherwise

        """
        start, end = start_end
        return minima[(minima > start) & (minima < end)].shape[0] == 1

    @classmethod
    def _rsp_rate(cls, extrema: np.array, sampling_rate: int, desired_length: int) -> np.array:
        """Compute continuous respiration rate from extrema values.

        Parameters
        ----------
        extrema: :class:`numpy.array`
            List of respiration extrema (peaks or troughs)
        sampling_rate : float
            Sampling rate of recorded data
        desired_length : int
            Desired length of the output signal


        Returns
        -------
        :class:`numpy.array`
            Respiration rate array interpolated to desired length

        """
        rsp_rate_raw = (sampling_rate * 60) / np.ediff1d(extrema)
        # remove last sample
        x_old = extrema[:-1]
        x_new = np.linspace(x_old[0], x_old[-1], desired_length)
        return nk.signal_interpolate(x_old, rsp_rate_raw, x_new, method="linear")



"""Sleep/Wake detection using the *Scripps Clinic Algorithm*."""

import numpy as np
import pandas as pd

from biopsykit.sleep.sleep_wake_detection.algorithms._base import _SleepWakeBase
from biopsykit.sleep.sleep_wake_detection.utils import rescore
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.array_handling import sanitize_input_1d
from biopsykit.utils.dtypes import SleepWakeDataFrame, _SleepWakeDataFrame


class ScrippsClinic(_SleepWakeBase):
    """Class representing the *Scripps Clinic Algorithm* for sleep/wake detection based on activity counts."""

    def __init__(self, **kwargs):
        """Class representing the *Scripps Clinic Algorithm* for sleep/wake detection based on activity counts.

        The *Scripps Clinic Algorithm* runs sleep wake detection on epoch level activity data. Epochs are 1 minute
        long and activity is represented by an activity index which comes from Actigraph data or from raw acceleration
        data converted into activity index data.

        Parameters
        ----------
        scale_factor : float
            scale factor to use for the predictions (default corresponds to scale factor optimized for use with
            the activity index, if other activity measures are desired the scale factor can be modified or optimized.).
            Default: 0.30


        References
        ----------
        add reference

        """
        scale_factor = kwargs.get("scale_factor", 0.30)
        self.scale_factor: float = scale_factor
        """scale factor to use for the predictions (default corresponds to scale factor optimized for use with the
        activity index, if other activity measures are desired the scale factor can be modified or optimized.).
        Default: 0.30
        """
        super().__init__(**kwargs)

    def fit(self, data: arr_t, **kwargs):  # noqa: ARG002
        """Fit sleep/wake detection algorithm to input data.

        .. note::
            Algorithms that do not have to (re)fit a ML model before sleep/wake prediction, such as rule-based
            algorithms, will internally bypass this method as the ``fit`` step is not needed.

        Parameters
        ----------
        data : array_like
            input data

        """
        return

    def predict(self, data: arr_t, **kwargs) -> SleepWakeDataFrame:
        """Apply sleep/wake prediction algorithm on input data.

        Parameters
        ----------
        data : array_like
            input data with activity index values
        **kwargs :
            additional arguments to be passed to the algorithm for prediction, such as:

            * ``rescore_data`` (``bool``):
              ``True`` to apply Webster's rescoring rules to the sleep/wake predictions, ``False`` otherwise.
              Default: ``True``
            * ``epoch_length`` (``int``):
              activity data epoch lengths in seconds, i.e. Epoch lengths are usually 30 or 60 seconds.
              Default: 60

        Returns
        -------
        :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`
            dataframe with sleep/wake predictions

        """
        index = None
        rescore_data: bool = kwargs.get("rescore_data", True)
        epoch_length: int = kwargs.get("epoch_length", 30)

        if isinstance(data, pd.DataFrame):
            index = data.index
            data = sanitize_input_1d(data)

        # ensure numpy
        # problem in my view: with np.convolve and "same" we dont make the "unsymmetric" convolution...
        # ( formula is with A-4 and A+2)
        # possible solution: np.valid and append 4 zeros in front and 2 zeros at the back or like implemented
        sf = np.array(self.scale_factor)
        kernel = sf * np.array(
            [
                0.0064,
                0.0074,
                0.0112,
                0.0112,
                0.0118,
                0.0118,
                0.0128,
                0.0188,
                0.0280,
                0.0664,
                0.0300,
                0.0112,
                0.100,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        scores = np.convolve(data, kernel, "same")

        scores[scores >= 1] = 99  # wake = 0
        scores[scores < 1] = 1  # sleep = 1
        scores[scores == 99] = 0  # wake = 0

        if rescore_data:
            scores = rescore(scores, epoch_length=epoch_length)

        if index is not None:
            scores = pd.DataFrame(scores, index=index, columns=["sleep_wake"])

        return _SleepWakeDataFrame(scores)



import warnings

import neurokit2 as nk
import numpy as np
import pandas as pd

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.ecg.event_extraction._base_ecg_extraction import BaseEcgExtraction
from biopsykit.utils.array_handling import sanitize_input_series
from biopsykit.utils.dtypes import (
    EcgRawDataFrame,
    HeartbeatSegmentationDataFrame,
    is_ecg_raw_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_q_peak_dataframe,
)
from biopsykit.utils.exceptions import EventExtractionError


class QPeakExtractionSciPyFindPeaksNeurokit(BaseEcgExtraction, CanHandleMissingEventsMixin):
    """Algorithm for Q-peak extraction using the ``scipy.find_peaks`` method implemented in NeuroKit2 [1]_.

    References
    ----------
    .. [1] Makowski, D., Pham, T., Lau, Z. J., Brammer, J. C., Lesspinasse, F., Pham, H., Schölzel, C., & S.H. Chen
        (2021). NeuroKit2: A Python Toolbox for Neurophysiological Signal Processing. Behavior Research Methods.
        https://doi.org/10.3758/s13428-020-01516-y

    """

    def __init__(self, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Initialize new QPeakExtractionSciPyFindPeaksNeurokit algorithm instance.

        Parameters
        ----------
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle missing data in the input dataframes. Default: "warn"

        """
        super().__init__(handle_missing_events=handle_missing_events)

    # @make_action_safe
    def extract(
        self,
        *,
        ecg: EcgRawDataFrame,
        heartbeats: HeartbeatSegmentationDataFrame,
        sampling_rate_hz: float,
    ):
        """Extract Q-peaks from given ECG signal.

        The results are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        ecg: :class:`~pandas.DataFrame`
            ECG signal
        heartbeats: :class:`~pandas.DataFrame`
            DataFrame containing one row per segmented heartbeat, each row contains start, end, and R-peak
            location (in samples from beginning of signal) of that heartbeat, index functions as id of heartbeat
        sampling_rate_hz: int
            Sampling rate of ECG signal in hz

        Returns
        -------
            self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        self._check_valid_missing_handling()
        is_ecg_raw_dataframe(ecg)
        is_heartbeat_segmentation_dataframe(heartbeats)
        ecg = sanitize_input_series(ecg, name="ecg")
        ecg = ecg.squeeze()

        # result df
        q_peaks = pd.DataFrame(index=heartbeats.index, columns=["q_peak", "nan_reason"])

        # used subsequently to store ids of heartbeats for which no AO or IVC could be detected
        heartbeats_no_q = []
        heartbeats_q_after_r = []

        # some neurokit functions (for example ecg_delineate()) don't work with r-peaks input as Series, so list instead
        r_peaks = list(heartbeats["r_peak_sample"])

        _, waves = nk.ecg_delineate(
            ecg, rpeaks=r_peaks, sampling_rate=int(sampling_rate_hz), method="peaks", show=False
        )

        extracted_q_peaks = waves["ECG_Q_Peaks"]

        # find heartbeat to which Q-peak belongs and save Q-peak position in corresponding row
        for idx, q in enumerate(extracted_q_peaks):
            # for some heartbeats, no Q can be detected, will be NaN in resulting df
            if np.isnan(q):
                heartbeats_no_q.append(idx)
            else:
                heartbeat_idx = heartbeats.loc[(heartbeats["start_sample"] < q) & (q < heartbeats["end_sample"])].index[
                    0
                ]

                # Q occurs after R, which is not valid
                if heartbeats["r_peak_sample"].loc[heartbeat_idx].item() < q:
                    heartbeats_q_after_r.append(heartbeat_idx)
                    q_peaks.loc[heartbeat_idx, "q_peak"] = np.NaN
                # valid Q-peak found
                else:
                    q_peaks.loc[heartbeat_idx, "q_peak"] = q

        # inform user about missing Q-values
        if q_peaks.isna().sum()[0] > 0:
            nan_rows = q_peaks[q_peaks["q_peak"].isna()]
            nan_rows = nan_rows.drop(index=heartbeats_q_after_r)
            nan_rows = nan_rows.drop(index=heartbeats_no_q)

            missing_str = f"No Q-peak detected in {q_peaks.isna().sum()[0]} heartbeats:\n"
            if len(heartbeats_no_q) > 0:
                q_peaks.loc[heartbeats_no_q, "nan_reason"] = "no_q_peak"
                missing_str += (
                    f"- for heartbeats {heartbeats_no_q} the neurokit algorithm was not able to detect a Q-peak\n"
                )
            if len(heartbeats_q_after_r) > 0:
                q_peaks.loc[heartbeats_no_q, "nan_reason"] = "q_after_r_peak"
                missing_str += (
                    f"- for heartbeats {heartbeats_q_after_r} the detected Q is invalid "
                    f"because it occurs after the R-peak\n"
                )
            if len(nan_rows.index.values) > 0:
                q_peaks.loc[nan_rows.index, "nan_reason"] = "no_q_peak_within_heartbeats"
                missing_str += (
                    f"- for {nan_rows.index.to_numpy()} apparently none of the found Q-peaks "
                    f"were within these heartbeats"
                )

            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)

        q_peaks = q_peaks.astype({"q_peak_sample": "Int64", "nan_reason": "object"})
        is_q_peak_dataframe(q_peaks)

        self.points_ = q_peaks
        return self



"""Functions to process sleep data from raw IMU data or Actigraph data."""

from collections.abc import Sequence
from typing import Any, Optional, Union

import numpy as np

from biopsykit.signals.imu import convert_acc_data_to_g
from biopsykit.signals.imu.activity_counts import ActivityCounts
from biopsykit.signals.imu.rest_periods import RestPeriods
from biopsykit.signals.imu.wear_detection import WearDetection
from biopsykit.sleep.sleep_endpoints import compute_sleep_endpoints
from biopsykit.sleep.sleep_wake_detection.sleep_wake_detection import SleepWakeDetection
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.array_handling import accumulate_array


def predict_pipeline_acceleration(
    data: arr_t, sampling_rate: float, convert_to_g: Optional[bool] = True, **kwargs
) -> dict[str, Any]:
    """Apply sleep processing pipeline on raw acceleration data.

    This function processes raw acceleration data collected during sleep. The pipeline consists of the following steps:

    * *Activity Count Conversion*: Convert (3-axis) raw acceleration data into activity counts. Most sleep/wake
      detection algorithms use activity counts (as typically provided by Actigraph devices) as input data.
    * *Wear Detection*: Detect wear and non-wear periods. Cut data to the longest continuous wear block.
    * *Rest Periods*: Detect rest periods, i.e., periods with large physical inactivity. The longest continuous
      rest period (*Major Rest Period*) is used to determine the *Bed Interval*, i.e., the period spent in bed.
    * *Sleep/Wake Detection*: Apply sleep/wake detection algorithm to classify phases of sleep and wake.
    * *Sleep Endpoint Computation*: Compute Sleep Endpoints from sleep/wake detection results and bed interval.

    Parameters
    ----------
    data : array_like with shape (n,3)
        input data. Must be a 3-d acceleration signal
    sampling_rate : float
        sampling rate of recorded data in Hz
    convert_to_g : bool, optional
        ``True`` if input data is provided in :math:`m/s^2` and should be converted in :math:`g`, ``False`` if input
        data is already in :math:`g` and does not need to be converted.
        Default: ``True``
    **kwargs :
        additional parameters to configure sleep/wake detection. The possible parameters depend on the selected
        sleep/wake detection algorithm and are passed to
        :class:`~biopsykit.sleep.sleep_wake_detection.SleepWakeDetection`. Examples are:

        * *algorithm_type*: name of sleep/wake detection algorithm to internally use for sleep/wake detection.
          Default: "Cole/Kripke"
        * *epoch_length*: epoch length in seconds. Default: 60


    Returns
    -------
    dict
        dictionary with Sleep Processing Pipeline results.

    """
    # TODO: add entries of result dictionary to docstring and add possibility to specify sleep/wake prediction algorithm
    kwargs.setdefault("algorithm_type", "cole_kripke")
    kwargs.setdefault("epoch_length", 60)
    ac = ActivityCounts(sampling_rate)
    wd = WearDetection(sampling_rate=sampling_rate)
    rp = RestPeriods(sampling_rate=sampling_rate)
    sw = SleepWakeDetection(**kwargs)

    if convert_to_g:
        data = convert_acc_data_to_g(data, inplace=False)

    df_wear = wd.predict(data)
    major_wear_block = wd.get_major_wear_block(df_wear)

    # cut data to major wear block
    data = wd.cut_to_wear_block(data, major_wear_block)

    if len(data) == 0:
        return {}

    df_ac = ac.calculate(data)
    df_ac = accumulate_array(df_ac, 1, 1 / kwargs.get("epoch_length"))

    df_sw = sw.predict(df_ac)
    df_rp = rp.predict(data)
    bed_interval = [df_rp["start"][0], df_rp["end"][0]]
    print(df_sw.value_counts())
    sleep_endpoints = compute_sleep_endpoints(df_sw, bed_interval)
    if not sleep_endpoints:
        return {}

    major_wear_block = [str(d) for d in major_wear_block]

    dict_result = {
        "wear_detection": df_wear,
        "activity_counts": df_ac,
        "sleep_wake_prediction": df_sw,
        "major_wear_block": major_wear_block,
        "rest_periods": df_rp,
        "bed_interval": bed_interval,
        "sleep_endpoints": sleep_endpoints,
    }
    return dict_result


def predict_pipeline_actigraph(
    data: arr_t, algorithm_type: str, bed_interval: Sequence[Union[str, int, np.datetime64]], **kwargs
) -> dict[str, Any]:
    """Apply sleep processing pipeline on actigraph data.

    This function processes actigraph data collected during sleep and performs sleep/wake detection.

    Parameters
    ----------
    data : array_like with shape (n,3)
        input data. Must be a 3-d acceleration signal
    algorithm_type : str
        name of sleep/wake detection algorithm to internally use for sleep/wake detection
    bed_interval : array_like
        beginning and end of bed interval, i.e., the time spent in bed

    **kwargs :
        additional parameters to configure sleep/wake detection. The possible parameters depend on the selected
        sleep/wake detection algorithm and are passed to
        :class:`~biopsykit.sleep.sleep_wake_detection.SleepWakeDetection`.


    Returns
    -------
    dict
        dictionary with Sleep Processing Pipeline results.

    """
    # TODO: add entries of result dictionary to docstring and add possibility to specify sleep/wake prediction algorithm
    sw = SleepWakeDetection(algorithm_type=algorithm_type, **kwargs)
    df_sw = sw.predict(data[["activity"]])

    sleep_endpoints = compute_sleep_endpoints(df_sw, bed_interval)

    dict_result = {"sleep_wake_prediction": df_sw, "sleep_endpoints": sleep_endpoints}
    return dict_result



import warnings

import numpy as np
import pandas as pd

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.icg.event_extraction._base_b_point_extraction import BaseBPointExtraction
from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.dtypes import (
    CPointDataFrame,
    HeartbeatSegmentationDataFrame,
    IcgRawDataFrame,
    is_b_point_dataframe,
    is_c_point_dataframe,
    is_heartbeat_segmentation_dataframe,
    is_icg_raw_dataframe,
)
from biopsykit.utils.exceptions import EventExtractionError

__all__ = ["BPointExtractionDrost2022"]


class BPointExtractionDrost2022(BaseBPointExtraction, CanHandleMissingEventsMixin):
    """B-point extraction algorithm by Drost et al. (2022) [1]_.

    This algorithm extracts B-points based on the maximum distance of the dZ/dt curve and a straight line fitted
    between the C-Point and the Point on the dZ/dt curve 150 ms before the C-Point.

    References
    ----------
    .. [1] Drost, L., Finke, J. B., Port, J., & Schächinger, H. (2022). Comparison of TWA and PEP as indices of a2- and
        ß-adrenergic activation. Psychopharmacology. https://doi.org/10.1007/s00213-022-06114-8

    """

    def __init__(self, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Initialize new ``BPointExtractionDrost2022`` instance.

        Parameters
        ----------
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle failing event extraction. Can be one of:
                * "warn": issue a warning and set the event to NaN
                * "raise": raise an ``EventExtractionError``
                * "ignore": ignore the error and continue with the next event
            Default: "warn"

        """
        super().__init__(handle_missing_events=handle_missing_events)

    # @make_action_safe
    def extract(
        self,
        *,
        icg: IcgRawDataFrame,
        heartbeats: HeartbeatSegmentationDataFrame,
        c_points: CPointDataFrame,
        sampling_rate_hz: float,
    ):
        """Extract B-points from given ICG derivative signal.

        This algorithm extracts B-points based on the maximum distance of the dZ/dt curve and a straight line
        fitted between the C-Point and the Point on the dZ/dt curve 150 ms before the C-Point.

        The results are saved in the ``points_`` attribute of the super class.

        Parameters
        ----------
        icg : :class:`~pandas.DataFrame`
            ICG derivative signal
        heartbeats : :class:`~pandas.DataFrame`
            Segmented heartbeats. Each row contains start, end, and R-peak location (in samples
            from beginning of signal) of that heartbeat, index functions as id of heartbeat
        c_points : :class:`~pandas.DataFrame`
            Extracted C-points. Each row contains the C-point location (in samples from beginning of signal) for each
            heartbeat, index functions as id of heartbeat. C-point locations can be NaN if no C-points were detected
            for certain heartbeats
        sampling_rate_hz : int
            sampling rate of ICG derivative signal in hz

        Returns
        -------
            self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        self._check_valid_missing_handling()
        is_icg_raw_dataframe(icg)
        is_heartbeat_segmentation_dataframe(heartbeats)
        is_c_point_dataframe(c_points)
        icg = sanitize_input_dataframe_1d(icg, column="icg_der")
        icg = icg.squeeze()

        # Create the b_point Dataframe. Use the heartbeats id as index
        b_points = pd.DataFrame(index=heartbeats.index, columns=["b_point_sample", "nan_reason"])

        # get the c_point locations from the c_points dataframe and search for entries containing NaN
        check_c_points = pd.isna(c_points["c_point_sample"]).to_numpy()

        # iterate over each heartbeat
        for idx, _data in heartbeats.iterrows():
            # check if c_points contain NaN. If this is the case, set the b_point to NaN and continue
            # with the next iteration
            if check_c_points[idx]:
                b_points.loc[idx, "b_point_sample"] = np.NaN
                b_points.loc[idx, "nan_reason"] = "c_point_nan"
                continue
            # Get the C-Point location at the current heartbeat id
            c_point = c_points.loc[idx, "c_point_sample"]

            # Calculate the start position of the straight line (150 ms before the C-Point) and ensure that the
            # start position is not negative
            line_start = max(c_point - int((150 / 1000) * sampling_rate_hz), 0)

            # Calculate the values of the straight line
            line_values = self._get_straight_line(line_start, icg.iloc[line_start], c_point, icg.iloc[c_point])

            # Get the interval of the cleaned ICG-signal in the range of the straight line
            signal_clean_interval = icg.iloc[line_start:c_point].squeeze()

            # Calculate the distance between the straight line and the cleaned ICG-signal
            distance = line_values["result"].to_numpy() - signal_clean_interval.to_numpy()

            # Calculate the location of the maximum distance and transform the index relative to the complete signal
            # to obtain the B-Point location
            b_point = line_start + np.argmax(distance)

            b_points.loc[idx, "b_point_sample"] = b_point

        num_nan = b_points["b_point_sample"].isna().sum()
        if num_nan > 0:
            idx_nan = b_points["b_point_sample"].isna()
            idx_nan = list(b_points.index[idx_nan])

            missing_str = (
                f"The C-point contains NaN at heartbeats {idx_nan}! The index of the B-points were also set to NaN."
            )
            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)

        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        is_b_point_dataframe(b_points)

        self.points_ = b_points
        return self

    @staticmethod
    def _get_straight_line(start_x: int, start_y: float, c_x: int, c_y: float):
        """Compute the values of a straight line fitted between the C-Point and the point 150 ms before the C-Point.

        Parameters
        ----------
        start_x: int
            index of the point 150 ms before the C-Point
        start_y: float
            value of the point 150 ms before the C-Point
        c_x: int
            index of the C-Point
        c_y: float
            value of the C-Point

        Returns
        -------
        :class:`~pandas.DataFrame`
            DataFrame containing the values of the straight line for each index between the C-Point and the point
            150 ms before the C-Point

        """
        # Compute the slope of the straight line
        start_y = float(start_y)
        c_y = float(c_y)
        slope = float((c_y - start_y) / (c_x - start_x))

        # Get the sample positions where we want to calculate the values of the straight line
        index = np.arange(0, (c_x - start_x), 1)
        line_values = pd.DataFrame(index=index, columns=["result"])

        # Compute the values of the straight line for each index
        line_values["result"] = (index * slope) + start_y

        return line_values



"""General class for sleep/wake detection."""

from typing import Optional

from biopsykit.sleep.sleep_wake_detection.algorithms.cole_kripke import ColeKripke
from biopsykit.sleep.sleep_wake_detection.algorithms.cole_kripke_old import ColeKripkeOld
from biopsykit.sleep.sleep_wake_detection.algorithms.sadeh import Sadeh
from biopsykit.sleep.sleep_wake_detection.algorithms.sazonov import Sazonov
from biopsykit.sleep.sleep_wake_detection.algorithms.scripps_clinic import ScrippsClinic
from biopsykit.sleep.sleep_wake_detection.algorithms.webster import Webster
from biopsykit.utils._types_internal import arr_t
from biopsykit.utils.dtypes import SleepWakeDataFrame, _SleepWakeDataFrame


class SleepWakeDetection:
    """General class for sleep/wake detection."""

    sleep_wake_algo = None

    def __init__(self, algorithm_type: Optional[str] = None, **kwargs):
        """General class for sleep/wake detection.

        This class provides a generalized interface for sleep/wake detection independent of the used algorithm.
        When initializing a new instance the algorithm type can be specified.>

        Parameters
        ----------
        algorithm_type : str, optional
            name of sleep/wake detection algorithm to internally use for sleep/wake detection or ``None`` to use
            default algorithm (Cole/Kripke Algorithm, see
            :class:`~biopsykit.sleep.sleep_wake_detection.algorithms.cole_kripke.ColeKripke` for further information)
        kwargs
            optional parameters to be passed to the sleep/wake detection algorithm. The possible parameters depend on
            the selected sleep/wake detection algorithm and are passed to the respective class.

        """
        available_sleep_wake_algorithms = {
            "cole_kripke": ColeKripke,
            "sadeh": Sadeh,
            "cole_kripke_old": ColeKripkeOld,
            "webster": Webster,
            "scripps_clinic": ScrippsClinic,
            "sazonov": Sazonov,
        }

        if algorithm_type is None:
            algorithm_type = "cole_kripke"

        if algorithm_type not in available_sleep_wake_algorithms:
            raise ValueError(
                f"Invalid algorithm type for sleep/wake detection! "
                f"Must be one of {available_sleep_wake_algorithms}, got {algorithm_type}."
            )

        sleep_wake_cls = available_sleep_wake_algorithms[algorithm_type]

        if (
            sleep_wake_cls is ColeKripke
            or sleep_wake_cls is ColeKripkeOld
            or sleep_wake_cls is Webster
            or sleep_wake_cls is ScrippsClinic
        ):
            self.sleep_wake_algo = sleep_wake_cls(**kwargs)

        else:
            self.sleep_wake_algo = sleep_wake_cls()

    def predict(self, data: arr_t, **kwargs) -> SleepWakeDataFrame:
        """Apply sleep/wake prediction on input data.

        Parameters
        ----------
        data : array_like
            input data
        **kwargs :
            additional arguments to be passed to the sleep/wake detection algorithm.
            The possible arguments depend on the individual algorithm.

        Returns
        -------
        :obj:`~biopsykit.utils.dtypes.SleepWakeDataFrame`
            dataframe with sleep/wake predictions

        """
        return _SleepWakeDataFrame(self.sleep_wake_algo.predict(data, **kwargs))



"""Functions to handle multicollinearity in data."""

from typing import Optional

import numpy as np
import pandas as pd


def remove_multicollinearity_correlation(data: pd.DataFrame, threshold: Optional[float] = 0.8) -> pd.DataFrame:
    """Remove features with multicollinearity based on cross-correlation coefficient.

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        Input data with features to check for multicollinearity.
    threshold : float, optional
        Cross-correlation coefficient threshold. Features with a correlation coefficient above this value will be
        removed. Default: 0.8

    Returns
    -------
    :class:`pandas.DataFrame`
        Dataframe without features with high multicollinearity.

    """
    corr_data = pd.DataFrame(np.triu(np.abs(data.corr())), columns=data.columns)

    multicoll_columns = np.logical_and(corr_data >= threshold, corr_data < 1.0).any()
    return data.loc[:, ~multicoll_columns]



"""Module containing helper functions to handle time data."""

import datetime
import re
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
import pytz

from biopsykit.utils._datatype_validation_helper import _assert_is_dtype
from biopsykit.utils._types_internal import path_t

tz = pytz.timezone("Europe/Berlin")
utc = pytz.timezone("UTC")

__all__ = [
    "check_tz_aware",
    "extract_time_from_filename",
    "get_time_from_date",
    "time_to_timedelta",
    "timedelta_to_time",
    "tz",
]


def check_tz_aware(data: pd.DataFrame) -> bool:
    """Check whether dataframe index is timezone-aware.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        dataframe with index to check


    Returns
    -------
    bool
        ``True`` if index of dataframe is a :class:`~pandas.DatetimeIndex` and index is timezone-aware,
        ``False`` otherwise

    """
    _assert_is_dtype(data, pd.DataFrame)
    return isinstance(data.index, pd.DatetimeIndex) and (data.index.tzinfo is not None)


def extract_time_from_filename(
    file_path: path_t, filename_pattern: str, date_pattern: Optional[str] = None
) -> datetime.datetime:
    """Extract time information from filename.

    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        filename
    filename_pattern : str
        regex string indicating how to extract time information from filename
    date_pattern : str, optional
        date format pattern or ``None`` to use default date format pattern ("%Y%m%d_%H_%M_%S").
        Default: ``None``

    Returns
    -------
    :class:`datetime.datetime`
        extracted date information

    """
    # ensure pathlib
    file_path = Path(file_path)
    if date_pattern is None:
        date_pattern = "%Y%m%d_%H_%M_%S"
    start_time = re.findall(filename_pattern, file_path.name)
    if len(start_time) == 0:
        raise ValueError(f"No valid string matching the pattern '{filename_pattern}' found in filename!")
    start_time = start_time[0]
    start_time = datetime.datetime.strptime(start_time, date_pattern)
    return pd.to_datetime(start_time)


def get_time_from_date(
    data: pd.Series,
    is_utc: Optional[bool] = False,
    tz_convert: Optional[bool] = False,
    timezone: Optional[Union[str, datetime.tzinfo]] = None,
) -> pd.Series:
    """Extract time information from series with date information.

    Some functions expect only time information (hour, minute, second, ...)
    without date information (year, month, day). This function can be used to extract only the relevant time
    information from the complete datetime data.


    Parameters
    ----------
    data : :class:`~pandas.Series`
        series with date information
    is_utc : bool, optional
        ``True`` if datetime is in UTC, ``False`` otherwise. Default: ``False``
    tz_convert : bool, optional
        ``True`` to convert datetime into correct timezone before extracting time information or
        ``False`` to localize datetime. Default: ``False``
    timezone : str or :class:`datetime.tzinfo`
        timezone the datetime objects are in or should be converted to.
        Default: ``None``, which defaults to time zone "Europe/Berlin"


    Returns
    -------
    :class:`~pandas.Series`
        pandas series with time information extracted from datetime

    """
    _assert_is_dtype(data, pd.Series)
    if timezone is None:
        timezone = tz
    data = pd.to_datetime(data, utc=is_utc)
    if tz_convert or pd.DatetimeIndex(data).tzinfo is not None:
        data = data.dt.tz_convert(timezone)
    else:
        data = data.dt.tz_localize(timezone)

    data = data - data.dt.normalize()
    return data


def time_to_timedelta(data: pd.Series) -> pd.Series:
    """Convert time information in a series into ``datetime.timedelta`` data.

    Parameters
    ----------
    data : :class:`~pandas.Series`
        series with time information


    Returns
    -------
    :class:`~pandas.Series`
        series with data converted into :class:`datetime.timedelta`

    """
    _assert_is_dtype(data, pd.Series)
    if np.issubdtype(data.dtype, np.timedelta64):
        # data is already a timedelta
        return data
    return pd.to_timedelta(data.astype(str))


def timedelta_to_time(data: pd.Series) -> pd.Series:
    """Convert ``datetime.timedelta`` data in a series ``datetime.time`` data.

    Parameters
    ----------
    data : :class:`~pandas.Series`
        series with data as :class:`datetime.timedelta`


    Returns
    -------
    :class:`~pandas.Series`
        series with data converted into :class:`datetime.time`

    """
    data_cpy = data.copy()
    # ensure pd.Timedelta
    data = data + pd.Timedelta("0h")
    # convert to datetime
    data = datetime.datetime.min + data.dt.to_pytimedelta()
    # convert to time
    data = [d.time() if d is not pd.NaT else None for d in data]
    data = pd.Series(np.array(data), index=data_cpy.index, name=data_cpy.name)
    return data



"""Functions to process sleep data from raw IMU data or Actigraph data."""

from biopsykit.sleep.sleep_processing_pipeline.sleep_processing_pipeline import (
    predict_pipeline_acceleration,
    predict_pipeline_actigraph,
)

__all__ = ["predict_pipeline_acceleration", "predict_pipeline_actigraph"]



import warnings

import neurokit2 as nk
import numpy as np
import pandas as pd
from tpcp import Parameter

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals.ecg.segmentation._base_segmentation import BaseHeartbeatSegmentation

__all__ = ["HeartbeatSegmentationNeurokit"]

from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.dtypes import (
    EcgRawDataFrame,
    HeartbeatSegmentationDataFrame,
    is_ecg_raw_dataframe,
    is_heartbeat_segmentation_dataframe,
)
from biopsykit.utils.exceptions import EventExtractionError


class HeartbeatSegmentationNeurokit(BaseHeartbeatSegmentation, CanHandleMissingEventsMixin):
    """Segment ECG signal into individual heartbeats based on NeuroKit2 [1]_.

    This algorithm segments the ECG signal into individual heartbeats based on the R-peaks detected by Neurokit's
    ``ecg_peaks`` function. The start of each heartbeat is determined based on the R-peak and the current RR-interval.

    References
    ----------
    .. [1] Makowski, D., Pham, T., Lau, Z. J., Brammer, J. C., Lesspinasse, F., Pham, H., Schölzel, C., & S.H. Chen
        (2021). NeuroKit2: A Python Toolbox for Neurophysiological Signal Processing. Behavior Research Methods.
        https://doi.org/10.3758/s13428-020-01516-y

    """

    _action_methods = "extract"

    # input parameters
    variable_length: Parameter[bool]
    start_factor: Parameter[float]
    r_peak_detection_method: Parameter[str]

    # result
    heartbeat_list_: HeartbeatSegmentationDataFrame

    def __init__(
        self,
        *,
        variable_length: bool = True,
        start_factor: float = 0.35,
        r_peak_detection_method: str = "neurokit",
        handle_missing_events: HANDLE_MISSING_EVENTS = "warn",
    ):
        """Initialize new ``HeartbeatSegmentationNeurokit`` algorithm instance.

        Parameters
        ----------
        variable_length : bool, optional
            ``True`` if extracted heartbeats should have variable length (depending on the current RR-interval) or
            ``False`` if extracted heartbeats should have fixed length (same length for all heartbeats, depending
            on the mean heartrate of the complete signal, 35% of mean heartrate in seconds before R-peak and 50%
            after r_peak, see :func:`neurokit2.ecg_segment` for details).
            For variable length heartbeats, the start of the next heartbeat follows directly after end of last
            (ends exclusive); For fixed length heartbeats, there might be spaces between heartbeat borders, or they
            might overlap. Default: ``True``
        start_factor : float, optional
            only needed if ``variable_length=True``. This parameter defines where the start border between heartbeats
            is set depending on the RR-interval to previous heartbeat. For example, ``start_factor=0.35`` means that
            the beat start is set at 35% of current RR-distance before the R-peak of the beat
        r_peak_detection_method : str, optional
            Method to detect R-peaks that is passed to :func:`neurokit2.ecg_peaks`. Default: "neurokit"
        handle_missing_events : one of {"warn", "raise", "ignore"}, optional
            How to handle missing data in the input dataframes. Default: "warn"

        """
        self.variable_length = variable_length
        self.start_factor = start_factor
        self.r_peak_detection_method = r_peak_detection_method
        super().__init__(handle_missing_events=handle_missing_events)

    # @make_action_safe
    def extract(  # noqa: PLR0915, PLR0912, C901
        self,
        *,
        ecg: EcgRawDataFrame,
        sampling_rate_hz: float,
    ):
        """Segment ECG signal into heartbeats.

        The function uses R-peak detection to segment the ECG signal into heartbeats. The start of each heartbeat is
        determined based on the R-peak and the current RR-interval.

        The results (start and end sample, R-peak sample, current RR-interval in samples and milliseconds) are saved
        in the ``heartbeat_list_`` attribute.

        Parameters
        ----------
        ecg : :class:`~pandas.Series` or :class:`~pandas.DataFrame`
            ECG signal
        sampling_rate_hz : int
            Sampling rate of ECG signal in hz

        Returns
        -------
            self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the event extraction fails and ``handle_missing`` is set to "raise"

        """
        is_ecg_raw_dataframe(ecg)
        ecg = sanitize_input_dataframe_1d(ecg, column="ECG")
        heartbeats = pd.DataFrame(
            columns=["start_sample", "end_sample", "r_peak_sample"],
        )
        heartbeats.index.name = "heartbeat_id"
        heartbeats = heartbeats.astype("Int64")

        if ecg.empty:
            is_heartbeat_segmentation_dataframe(heartbeats)
            self.heartbeat_list_ = heartbeats
            missing_str = "No ECG signal found, no heartbeats can be segmented!"
            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)
            return self

        _, r_peaks = nk.ecg_peaks(ecg, sampling_rate=int(sampling_rate_hz), method=self.r_peak_detection_method)
        r_peaks = r_peaks["ECG_R_Peaks"]

        if len(r_peaks) < 2:
            # no r-peaks were detected, so no heartbeats can be segmented
            # clear dataframe
            is_heartbeat_segmentation_dataframe(heartbeats)
            missing_str = "Not sufficient R-peaks were detected, so no heartbeats can be segmented!"
            self.heartbeat_list_ = heartbeats
            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)
            return self

        heartbeats = pd.DataFrame(
            index=pd.Index(np.arange(0, len(r_peaks)), name="heartbeat_id"),
            columns=["start_sample", "end_sample", "r_peak_sample"],
        )

        heartbeats = heartbeats.assign(r_peak_sample=r_peaks)
        # save RR-interval to successive heartbeat
        heartbeats = heartbeats.assign(rr_interval_sample=np.abs(heartbeats["r_peak_sample"].diff(periods=-1)))

        if self.variable_length:
            # split ecg signal into heartbeats with varying length
            rr_interval_samples = heartbeats["r_peak_sample"].diff()

            # calculate start of each heartbeat based on corresponding R-peak and current RR-interval
            beat_starts = heartbeats["r_peak_sample"] - self.start_factor * rr_interval_samples

            # extrapolate first beats start based on RR-interval of next beat
            first_beat_start = heartbeats["r_peak_sample"].iloc[0] - self.start_factor * rr_interval_samples.iloc[1]
            if first_beat_start >= 0:
                beat_starts.iloc[0] = first_beat_start
            else:
                beat_starts = beat_starts.iloc[1:].reset_index(drop=True)  # drop row if heartbeat is incomplete
                heartbeats = heartbeats.iloc[1:].reset_index(drop=True)
            beat_starts = round(beat_starts).astype(int)

            # calculate beat ends (last beat ends 1 sample before next starts, end is exclusive)
            beat_ends = beat_starts.shift(-1)  # end is exclusive

            # extrapolate last beats end based on RR-interval of previous beat
            last_beat_end = round(
                heartbeats["r_peak_sample"].iloc[-1] + (1 - self.start_factor) * rr_interval_samples.iloc[-1]
            )

            if last_beat_end < len(ecg):
                beat_ends.iloc[-1] = last_beat_end
            else:
                # drop the last beat if it is incomplete
                heartbeats = heartbeats.iloc[:-1]
                beat_ends = beat_ends.iloc[:-1]
                beat_starts = beat_starts.iloc[:-1]
            beat_ends = beat_ends.astype(int)

            # extract time of each beat's start
            beat_starts_time = ecg.iloc[beat_starts].index
            heartbeats = heartbeats.assign(start_sample=beat_starts, end_sample=beat_ends, start_time=beat_starts_time)

        else:
            # split ecg signal into heartbeats with fixed length
            heartbeat_segments = nk.ecg_segment(ecg, rpeaks=r_peaks, sampling_rate=int(sampling_rate_hz), show=False)

            heartbeat_segments_new = {int(k) - 1: v for k, v in heartbeat_segments.items()}
            heartbeat_segments_new = pd.concat(heartbeat_segments_new, names=["heartbeat_id"])

            heartbeat_segments_new = heartbeat_segments_new.groupby("heartbeat_id").agg(
                start_sample=("Index", "first"),
                end_sample=("Index", "last"),
                start_time=("Index", lambda s: ecg.index[s.iloc[0]]),
            )
            # fill the empty columns of heartbeats with the start_sample, end_sample, and start_time of
            # heartbeat_segments_new
            heartbeats.update(heartbeat_segments_new)
            heartbeats = heartbeats.assign(start_time=heartbeat_segments_new["start_time"])

        # check if R-peak occurs between corresponding start and end
        check = heartbeats.apply(lambda x: x["start_sample"] < x["r_peak_sample"] < x["end_sample"], axis=1)
        if len(check.loc[~check]) > 0:
            raise ValueError(
                f"Start/end/R-peak position of heartbeat {list(check.loc[check is False].index)} could be incorrect!"
            )

        # ensure that index is Int64Index (not RangeIndex) because some neurokit functions won't work  with RangeIndex
        heartbeats.index = list(heartbeats.index)
        heartbeats.index.name = "heartbeat_id"

        heartbeats = heartbeats.assign(rr_interval_ms=heartbeats["rr_interval_sample"] / sampling_rate_hz * 1000)

        # ensure correct column order
        heartbeats = heartbeats[
            ["start_time", "start_sample", "end_sample", "r_peak_sample", "rr_interval_sample", "rr_interval_ms"]
        ]

        heartbeats = heartbeats.astype(
            {
                "start_sample": "Int64",
                "end_sample": "Int64",
                "r_peak_sample": "Int64",
                "rr_interval_sample": "Int64",
                "rr_interval_ms": "Float64",
            }
        )

        is_heartbeat_segmentation_dataframe(heartbeats)

        self.heartbeat_list_ = heartbeats
        return self
